import importlib
import pkgutil

import cauliflow.plugins as plugins


# https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/
def _iter_namespace(ns_pkg):
    # Specifying the second argument (prefix) to iter_modules makes the
    # returned name an absolute name instead of a relative one. This allows
    # import_module to work without having to do additional modification to
    # the name.
    return pkgutil.iter_modules(ns_pkg.__path__, ns_pkg.__name__ + ".")


class PluginManager:
    def __init__(self):
        pass

    def init(self):
        self._load_buildin()

    def _load_buildin(self):
        for finder, name, ispkg in _iter_namespace(plugins):
            importlib.import_module(name)



import asyncio
import subprocess
import sys
from pathlib import Path

import pytest
from aioca import caget, caput, purge_channel_caches

from cauliflow.context import ctx_blackboard, ctx_flowdata
from cauliflow.plugins.ca import CagetNode, CamonitorNode, CaputNode
from cauliflow.plugins.message import MessageNode

SOFT_RECORDS = str(Path(__file__).parent / "ca_records.db")
PREFIX = "ET_SASAKI"


@pytest.fixture
def ioc():
    process = subprocess.Popen(
        [
            sys.executable,
            "-m",
            "epicscorelibs.ioc",
            "-m",
            f"head={PREFIX}",
            "-d",
            SOFT_RECORDS,
        ],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        universal_newlines=True,
    )
    yield process
    purge_channel_caches()
    try:
        process.communicate("exit")
    except ValueError:
        pass


@pytest.mark.asyncio
async def test_camonitor(ioc, init_context_vars):
    blackboard = ctx_blackboard.get()
    blackboard["msg"] = []

    pvname = f"{PREFIX}:LONGOUT"
    node = CamonitorNode(name="ca", param_dict={"pvname": pvname})
    msg = MessageNode(
        name="msg", param_dict={"msg": "{{ bb.msg + [fd.ca] }}", "out_bb": True}
    )
    node.add_child(msg)

    async def caput_test():
        await asyncio.sleep(2.0)
        await caput(pvname, 5.0)
        await asyncio.sleep(0.1)
        await caput(pvname, 6.0)

    try:
        async with asyncio.timeout(5):
            async with asyncio.TaskGroup() as tg:
                tg.create_task(node.run())
                tg.create_task(caput_test())
    except TimeoutError:
        pass

    assert len(blackboard["msg"]) == 3
    assert blackboard["msg"][1]["value"] == 5.0
    assert blackboard["msg"][2]["value"] == 6.0


@pytest.mark.asyncio
async def test_caget_ok(ioc, init_context_vars):
    pvname = f"{PREFIX}:LONGOUT"
    node = CagetNode(name="node", param_dict={"pvname": pvname})

    await caput(pvname, 1)
    await node.run()

    flowdata = ctx_flowdata.get()
    data = flowdata["node"]

    assert data["name"] == pvname
    assert data["value"] == 1.0
    assert data["status"] == 0
    assert data["severity"] == 0
    assert data["ok"] is True


@pytest.mark.asyncio
async def test_caget_ok_list(ioc, init_context_vars):
    pvlist = [f"{PREFIX}:LONGOUT", f"{PREFIX}:AO"]
    node = CagetNode(name="node", param_dict={"pvname": pvlist})

    await caput(pvlist[0], 2)
    await caput(pvlist[1], 2)
    await node.run()

    flowdata = ctx_flowdata.get()
    data = flowdata["node"]

    assert len(data) == 2
    for d, pv in zip(data, pvlist):
        assert d["name"] == pv
        assert d["value"] == 2.0
        assert d["status"] == 0
        assert d["severity"] == 0
        assert d["ok"] is True


@pytest.mark.asyncio
async def test_caget_ng(ioc, init_context_vars):
    pvname = "JOHN_DOE"
    node = CagetNode(name="node", param_dict={"pvname": pvname, "timeout": 0.1})

    await node.run()

    flowdata = ctx_flowdata.get()
    data = flowdata["node"]

    assert data["name"] == pvname
    assert data["ok"] is False


@pytest.mark.asyncio
async def test_caput_ok(ioc, init_context_vars):
    pvname = f"{PREFIX}:LONGOUT"
    val = 10
    node = CaputNode(name="node", param_dict={"pvname": pvname, "value": val})

    await node.run()

    flowdata = ctx_flowdata.get()
    data = flowdata["node"]

    assert data["name"] == pvname
    assert data["ok"] is True

    get_val = await caget(pvname)

    assert get_val == val


@pytest.mark.asyncio
async def test_caput_list(ioc, init_context_vars):
    pvlist = [f"{PREFIX}:LONGOUT", f"{PREFIX}:AO"]
    values = [1, 2]
    node = CaputNode(name="node", param_dict={"pvname": pvlist, "value": values})

    await node.run()

    get_vals = await caget(pvlist)

    flowdata = ctx_flowdata.get()
    data = flowdata["node"]

    assert len(data) == 2
    for d, pv, name, v in zip(data, get_vals, pvlist, values):
        assert d["name"] == name
        assert d["ok"] is True
        assert pv == v


@pytest.mark.asyncio
async def test_caput_list_with_repeat(ioc, init_context_vars):
    pvlist = [f"{PREFIX}:LONGOUT", f"{PREFIX}:AO"]
    value = 10
    node = CaputNode(
        name="node",
        param_dict={"pvname": pvlist, "value": value, "repeat_value": True},
    )

    await node.run()

    get_vals = await caget(pvlist)

    flowdata = ctx_flowdata.get()
    data = flowdata["node"]

    assert len(data) == 2
    for d, pv, name in zip(data, get_vals, pvlist):
        assert d["name"] == name
        assert d["ok"] is True
        assert pv == value


@pytest.mark.asyncio
async def test_caput_ng(ioc, init_context_vars):
    pvname = "JOHN_DOE"
    val = 10
    node = CaputNode(
        name="node", param_dict={"pvname": pvname, "value": val, "timeout": 1}
    )

    await node.run()

    flowdata = ctx_flowdata.get()
    data = flowdata["node"]

    assert data["name"] == pvname
    assert data["ok"] is False



import sys
from pathlib import Path

sys.path.append(str(Path("_ext").resolve()))

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "cauliflow"
copyright = "2025, Shinya Sasaki"
author = "Shinya Sasaki"
release = "0.1.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


extensions = [
    "myst_parser",
    "sphinxcontrib.mermaid",
    "sphinx_multiversion",
    "cauliflow_node",
    "cauliflow_filters",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}

myst_heading_anchors = 3

html_context = {
    "display_github": True,
    "github_user": "sasaki77",
    "github_repo": "cauliflow",
    "github_version": "master/docs/",
}

github_doc_root = "https://github.com/sasaki77/cauliflow/master/doc/"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_theme_path = ["sphinx_rtd_theme.get_html_theme_path()"]

html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]

# sphinx-multiversion settings
smv_tag_whitelist = r"^\d+\.\d+.*$"
smv_branch_whitelist = "master"



import pytest

from cauliflow.context import ctx_blackboard, ctx_flowdata
from cauliflow.flow import ConcurrentFlows, Flow, SequentialFlows


@pytest.mark.asyncio
async def test_flow(init_plugins):
    flow = Flow("test")

    args1 = {"a": 1, "b": 1}
    args2 = {"a": "{{ fd['add1'] }}", "b": 1}
    args3 = {"a": "{{ fd['add2'] }}", "b": 1}

    flow.create_node("test.addnode", "root", "add1", args1)
    flow.create_node("test.addnode", "add1", "add2", args2)
    flow.create_node("test.addnode", "add2", "add3", args3)

    await flow.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["add1"] == 2
    assert flowdata["add2"] == 3
    assert flowdata["add3"] == 4


@pytest.mark.asyncio
async def test_sequential_flows(init_plugins):
    flow1 = Flow("test")
    args1 = {"a": 1, "b": 1, "out_bb": True}
    flow1.create_node("test.addnode", "root", "add1", args1)

    flow2 = Flow("test")
    args2 = {"a": "{{ bb['add1']}}", "b": 1}
    flow2.create_node("test.addnode", "root", "add2", args2)

    flows = SequentialFlows()
    flows.append(flow1)
    flows.append(flow2)

    await flows.run()
    flowdata = ctx_flowdata.get()
    blackboard = ctx_blackboard.get()
    assert blackboard["add1"] == 2
    assert flowdata["add2"] == 3


@pytest.mark.asyncio
async def test_concurrent_flows(init_plugins):
    flow1 = Flow("test")
    args1 = {"a": 1, "b": 1, "out_bb": True}
    flow1.create_node("test.addnode", "root", "add1", args1)

    flow2 = Flow("test")
    args2 = {"a": 1, "b": 1, "out_bb": True}
    flow2.create_node("test.addnode", "root", "add2", args2)

    flows = ConcurrentFlows()
    flows.append(flow1)
    flows.append(flow2)

    await flows.run()
    blackboard = ctx_blackboard.get()
    assert blackboard["add1"] == 2
    assert blackboard["add2"] == 2


@pytest.mark.asyncio
async def test_composite_flows(init_plugins):
    flow1 = Flow("test")
    args1 = {"a": 1, "b": 1, "out_bb": True}
    flow1.create_node("test.addnode", "root", "add1", args1)

    flow2 = Flow("test")
    args2 = {"a": "{{ bb['add1']}}", "b": 1, "out_bb": True}
    flow2.create_node("test.addnode", "root", "add2", args2)

    flow3 = Flow("test")
    args3 = {"a": 1, "b": 1, "out_bb": True}
    flow3.create_node("test.addnode", "root", "add3", args3)

    flow4 = Flow("test")
    args4 = {"a": "{{ bb['add3']}}", "b": 1, "out_bb": True}
    flow4.create_node("test.addnode", "root", "add4", args4)

    flows1 = SequentialFlows()
    flows1.append(flow1)
    flows1.append(flow2)

    flows2 = SequentialFlows()
    flows2.append(flow3)
    flows2.append(flow4)

    cflows = ConcurrentFlows()
    cflows.append(flows1)
    cflows.append(flows2)

    await cflows.run()
    blackboard = ctx_blackboard.get()
    assert blackboard["add2"] == 3
    assert blackboard["add4"] == 3


@pytest.mark.asyncio
async def test_flow_name_collision(init_plugins):
    flow = Flow("test")

    args1 = {"a": 1, "b": 1}
    args2 = {"a": "{{ fd['add1'] }}", "b": 1}
    args3 = {"a": "{{ fd['add2'] }}", "b": 1}

    flow.create_node("test.addnode", "root", "add1", args1)
    flow.create_node("test.addnode", "add1", "add2", args2)

    with pytest.raises(KeyError):
        flow.create_node("test.addnode", "add2", "add1", args3)



import pytest

from cauliflow.context import ctx_flowdata
from cauliflow.plugins.itemloop import ForDict, ForList


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "lists, expression, filter, expected",
    [
        ([[1, 2], [3, 4]], "item0*item1", "item0*item1>7", [3, 4, 6]),
        (
            [["head1", "head2"], {"name1": "val1", "name2": "val2"}],
            "item0+':'+item1_val",
            None,
            ["head1:val1", "head1:val2", "head2:val1", "head2:val2"],
        ),
        ([1, 2, 3, 4], "item0*2", "item0*2>7", [2, 4, 6]),
        (
            {"key1": 1, "key2": 2},
            "item0_key + '-' + item0_val | str",
            None,
            ["key1-1", "key2-2"],
        ),
        (
            [
                [
                    {"first": "hello, ", "second": "world"},
                    {"first": "foo", "second": "bar"},
                ]
            ],
            "item0.first + item0.second",
            None,
            ["hello, world", "foobar"],
        ),
    ],
)
async def test_for_list(init_context_vars, lists, expression, filter, expected):
    params = {"lists": lists, "expression": expression, "filter": filter}
    node = ForList(name="node", param_dict=params)
    await node.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["node"] == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "lists, key, val, filter, expected",
    [
        (
            [["head1", "head2"], {"name1": "val1", "name2": "val2"}],
            "item0 + ':' + item1_key",
            "item1_val",
            None,
            {
                "head1:name1": "val1",
                "head1:name2": "val2",
                "head2:name1": "val1",
                "head2:name2": "val2",
            },
        ),
        (
            [1, 2, 3, 4],
            "'key' + item0 |str",
            "item0*2",
            "item0*2>7",
            {"key1": 2, "key2": 4, "key3": 6},
        ),
        (
            {"key1": 1, "key2": 2},
            "item0_key + '-' + item0_val | str",
            "item0_val",
            None,
            {"key1-1": 1, "key2-2": 2},
        ),
    ],
)
async def test_for_dict(init_context_vars, lists, key, val, filter, expected):
    params = {"lists": lists, "key": key, "val": val, "filter": filter}
    node = ForDict(name="node", param_dict=params)
    await node.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["node"] == expected



from pprint import pprint

from cauliflow.context import ctx_flowdata
from cauliflow.node import ArgSpec, ProcessNode, node


@node.register("stdout")
class OutNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Print data to stdout.
      description:
        - stdout node prints the data to the standart output
      parameters:
        src:
          description:
            - Data to print.
            - If src is not set, the stdout node prints the flowdata.
        pretty:
          description:
            - Enable pretty print.

    EXAMPLE: |-
      # Print flowdata to stdout
      - stdout:
          name: "stdout"

      # Print "hello" to stdout
      - stdout:
          name: "stdout"
          src: "hello"

      # Pretty print the value of 'foo' field in the blackboard to stdout
      - stdout:
          name: "stdout"
          src: "{{ bb.foo }}"
          pretty: yes
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "src": ArgSpec(type="any", required=False, default=None),
            "pretty": ArgSpec(type="bool", required=False, default=False),
        }

    async def process(self) -> None:
        print_func = pprint if self.params["pretty"] else print
        src = self.params["src"]

        if src is None:
            fd = ctx_flowdata.get()
            print_func(fd)
            return

        print_func(src)



from collections import UserDict


class FlowData(UserDict):
    def __setitem__(self, key, item):
        if key in self.data:
            raise KeyError(f"Key '{key}' already exists. Overwriting is not allowed.")
        self.data[key] = item



import asyncio

import pytest

from cauliflow.context import ctx_blackboard
from cauliflow.plugins.buffer import BufferNode
from cauliflow.plugins.message import MessageNode


@pytest.mark.asyncio
async def test_buffer_full(init_context_vars):
    node = BufferNode(
        name="buffer",
        param_dict={
            "size": 5,
            "timeout": 5,
            "input": 1,
        },
    )
    msg = MessageNode(name="msg", param_dict={"msg": "{{ fd.buffer }}", "out_bb": True})
    node.add_child(msg)

    for _ in range(5):
        await node.run()

    blackboard = ctx_blackboard.get()
    assert blackboard["msg"] == [1, 1, 1, 1, 1]


@pytest.mark.asyncio
async def test_buffer_flatten(init_context_vars):
    node = BufferNode(
        name="buffer",
        param_dict={"size": 5, "timeout": 5, "input": [1, 1], "flatten": True},
    )
    msg = MessageNode(name="msg", param_dict={"msg": "{{ fd.buffer }}", "out_bb": True})
    node.add_child(msg)

    for _ in range(3):
        await node.run()

    blackboard = ctx_blackboard.get()
    assert blackboard["msg"] == [1, 1, 1, 1, 1, 1]


@pytest.mark.asyncio
async def test_buffer_timeout(init_context_vars):
    node = BufferNode(
        name="buffer",
        param_dict={
            "size": 5,
            "timeout": 1,
            "input": 1,
        },
    )
    msg = MessageNode(name="msg", param_dict={"msg": "{{ fd.buffer }}", "out_bb": True})
    node.add_child(msg)

    for _ in range(3):
        await node.run()
    await asyncio.sleep(2)

    blackboard = ctx_blackboard.get()
    assert blackboard["msg"] == [1, 1, 1]


@pytest.mark.asyncio
async def test_buffer_not_call_before_timeout(init_context_vars):
    node = BufferNode(
        name="buffer",
        param_dict={
            "size": 5,
            "timeout": 2,
            "input": 1,
        },
    )
    msg = MessageNode(name="msg", param_dict={"msg": "{{ fd.buffer }}", "out_bb": True})
    node.add_child(msg)

    for _ in range(3):
        await node.run()

    blackboard = ctx_blackboard.get()
    assert "msg" not in blackboard



import pytest

from cauliflow.blackboard import BlackBoard
from cauliflow.context import ctx_blackboard, ctx_flowdata, ctx_macros
from cauliflow.flowdata import FlowData
from cauliflow.macros import Macros
from cauliflow.node import ArgSpec, Node, node
from cauliflow.plugin_manager import PluginManager


@node.register("test.addnode")
class AddNode(Node):
    async def process(self) -> None:
        outbb = self.params["out_bb"]
        _sum = self.params["a"] + self.params["b"]
        if outbb:
            bb = ctx_blackboard.get()
            bb[self.name] = _sum
        else:
            flowdata = ctx_flowdata.get()
            flowdata[self.name] = _sum

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "a": ArgSpec(type="int", required=True),
            "b": ArgSpec(type="int", required=True),
            "out_bb": ArgSpec(type="bool", required=False, default=False),
        }


@pytest.fixture
def init_context_vars():
    bb = BlackBoard()
    fd = FlowData()
    macros = Macros()

    ctx_blackboard.set(bb)
    ctx_flowdata.set(fd)
    ctx_macros.set(macros)


@pytest.fixture(scope="module")
def init_plugins():
    pm = PluginManager()
    pm.init()



import pytest

from cauliflow.plugins.file import OutFileNode


@pytest.mark.asyncio
async def test_output_file(init_context_vars, tmpdir_factory):
    fn = tmpdir_factory.mktemp("data").join("test.txt")
    with open(fn, mode="w") as f:
        f.write("hello\n")

    node = OutFileNode(name="node", param_dict={"path": str(fn), "src": "hello"})
    await node.run()
    text = ""
    with open(fn) as f:
        text = f.read()
    assert text == "hello\nhello\n"

    await node.run()
    text = ""
    with open(fn) as f:
        text = f.read()
    assert text == "hello\nhello\nhello\n"



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any

from cauliflow.context import (
    ContextNode,
    ctx_blackboard,
    ctx_flowdata,
    ctx_flows,
    ctx_node,
)
from cauliflow.logging import get_logger
from cauliflow.variable import Variable


@dataclass(frozen=True)
class ArgSpec:
    type: str
    required: bool = True
    default: Any = None


_logger = get_logger(__name__)

COMMON_ARGUMENT_SPEC: dict[str, ArgSpec] = {
    "out_bb": ArgSpec(type="bool", required=False, default=False),
    "out_field": ArgSpec(type="str", required=False, default=None),
}

COMMON_ARGUMENT_DOC = """
out_bb:
  description:
    - If this parameter is set to True, the data is stored in the blackboard rather than in the flowdata.
    - By default, the key in the blackboard is set to the node name. To customize the key, specify the out_field parameter.
out_field:
  description:
    - The key for the data in the blackboard. If out_bb is set to False, this parameter is ignored.
"""


class Node(ABC):
    DOCUMENTATION = None
    EXAMPLES = None

    def __init__(self, name: str, param_dict: dict | None = None):
        self.name: str = name
        self.child: Node | None = None
        self.enable_output = False
        self.argument_spec: dict[str, ArgSpec] = {}
        self.argument_spec.update(self.set_argument_spec())
        self.vars = None
        self.params = {}

        if param_dict is not None:
            self.vars = self._make_vars(self.argument_spec, param_dict)

    @abstractmethod
    async def process(self) -> None: ...

    async def run(self) -> None:
        await self._run_self()
        await self._run_child()

    async def _run_self(self) -> None:
        ctx_node.set(ContextNode(name=self.name))
        self._fetch_params()
        await self.process()
        flows = ctx_flows.get()
        if flows.debug:
            _log_debug()

    async def _run_child(self) -> None:
        if self.child is None:
            return
        await self.child.run()

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {}

    def set_common_output_args(self) -> None:
        self.enable_output = True
        self.argument_spec.update(COMMON_ARGUMENT_SPEC)

    def set_params(self, params: dict) -> None:
        self.vars = self._make_vars(self.argument_spec, params)

    def add_child(self, child: "Node", param: str | None = None) -> None:
        self.child = child

    def output(self, value: Any) -> None:
        if self.enable_output is False:
            _logger.warning(
                "output is disabled. call set_common_output_args in set_argumet_spec method"
            )
            return

        param_out_field = self.params["out_field"]
        field = param_out_field if param_out_field else self.name

        ctx = ctx_blackboard if self.params["out_bb"] else ctx_flowdata

        var = ctx.get()
        var[field] = value

    def _make_vars(
        self, argument_spec: dict[str, ArgSpec], param_dict: dict
    ) -> dict[str, Variable]:
        vars = {}

        for k, v in argument_spec.items():
            if k in param_dict:
                vars[k] = Variable(param_dict[k])
                continue

            if v.required is True:
                msg = f"missing parameter required by {k}"
                raise TypeError(msg)

            vars[k] = Variable(v.default)

        return vars

    def _fetch_params(self) -> None:
        if self.vars is None:
            raise TypeError("vars is not initialized")

        for k, v in self.vars.items():
            val = v.fetch()
            self.params[k] = val


class TriggerNode(Node):
    pass


class ProcessNode(Node):
    pass


class FlowControlNode(Node):
    pass


def _log_debug() -> None:
    flowdata = ctx_flowdata.get()
    blackboard = ctx_blackboard.get()
    _logger.debug(f"flowdata={flowdata}, blackboard={blackboard}")


class NodeFactory:
    registry = {}

    @classmethod
    def create(cls, _name: str, name: str, param_dict: dict) -> Node:
        node_class = cls.registry[_name]
        node = node_class(name=name, param_dict=param_dict)
        return node

    @classmethod
    def register(cls, name: str):
        _logger.debug(f"{name} is registered")

        def inner_wrapper(wrapped_class: type[Node]):
            if name in cls.registry:
                _logger.warning(f"{name} is already registered.")
            cls.registry[name] = wrapped_class
            return wrapped_class

        return inner_wrapper

    @classmethod
    def get(cls, name: str) -> Node:
        if name not in cls.registry:
            raise KeyError
        return cls.registry[name]


node = NodeFactory



from cauliflow.filters import dict2item, dict_keys, dict_values, join


def test_dict_keys():
    item = dict_keys({"key1": 1, "key2": 2})
    assert item == ["key1", "key2"]


def test_dict_values():
    item = dict_values({"key1": 1, "key2": 2})
    assert item == [1, 2]


def test_dict2item():
    item = dict_values({"key1": 1, "key2": 2})
    item = dict2item("k", "v", {"key1": 1, "key2": 2})
    assert item == [{"k": "key1", "v": 1}, {"k": "key2", "v": 2}]


def test_join():
    item = join(", ", ["Hello", "world"])
    assert item == "Hello, world"



import asyncio
from time import monotonic

from apscheduler.schedulers.asyncio import AsyncIOScheduler
from apscheduler.triggers.cron import CronTrigger

from cauliflow.context import init_flowdata
from cauliflow.node import ArgSpec, TriggerNode, node


@node.register("interval")
class IntervalNode(TriggerNode):
    """
    DOCUMENTATION:
      short_description: Run child node at a regular interval.
      description:
        - Run child node at a regular interval
      parameters:
        interval:
          description:
            - Interval to run child node in second.
    EXAMPLE: |-
      # Run child node every second
      - interval:
          name: "interval"
          interval: 1.0

      # Run child node. Interval is set by macro.
      - interval:
          name: "interval"
          interval: "{{ macro.interval }}"
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {"interval": ArgSpec(type="float", required=True)}

    async def process(self) -> None:
        interval = self.params["interval"]
        next_time = monotonic()

        while True:
            init_flowdata()
            if self.child is not None:
                await self.child.run()
            next_time += interval
            sleep_duration = max(0, next_time - monotonic())
            await asyncio.sleep(sleep_duration)


@node.register("scheduler")
class Scheduler(TriggerNode):
    """
    DOCUMENTATION:
      short_description: Run child node at a cron schedule.
      description:
        - Run child node at a cron schedule.
      parameters:
        cron:
          description:
            - Specify schedule in cron style.
    EXAMPLE: |-
      # Run child node every minute
      - scheduler:
          name: "scheduler"
          cron: "*/1 * * * *"

      # Run child node every day at 12:30
      - scheduler:
          name: "scheduler"
          cron: "30 12 * * *"
    """

    _scheduler = AsyncIOScheduler()

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "cron": ArgSpec(type="str", required=True),
        }

    async def process(self) -> None:
        cron = self.params["cron"]

        cron_args = _parse_cron_string(cron)
        trigger = CronTrigger(**cron_args)
        self._scheduler.add_job(self._job, trigger)

        if not self._scheduler.running:
            self._scheduler.start()

        while True:
            await asyncio.sleep(1)

    async def _job(self):
        init_flowdata()
        if self.child is not None:
            await self.child.run()


def _parse_cron_string(cron_string: str):
    parts = cron_string.split()
    if len(parts) != 5:
        raise ValueError(
            "Invalid cron string format. Expected format: 'minute hour day month day_of_week'"
        )

    return {
        "minute": parts[0],
        "hour": parts[1],
        "day": parts[2],
        "month": parts[3],
        "day_of_week": parts[4],
    }



from enum import StrEnum
from logging import getLogger
from pathlib import Path

import yaml
from cauliflow.context import ctx_macros
from cauliflow.flow import ConcurrentFlows, Flow, Flows, SequentialFlows
from cauliflow.macros import Macros

_logger = getLogger(__name__)


class FlowsType(StrEnum):
    SEQUENCTIAL = "sequential"
    CONCURRENT = "concurrent"


def flow_from_yaml(file_path: str | Path) -> Flows:
    yaml_dict = _load_yaml(file_path)

    is_seq = FlowsType.SEQUENCTIAL in yaml_dict
    is_con = FlowsType.CONCURRENT in yaml_dict
    is_onlyflow = "flow" in yaml_dict

    if (is_seq + is_con + is_onlyflow) > 1:
        _logger.error("multiple flows are detected")

    flows = None
    if is_seq:
        flows = _make_seq(yaml_dict[FlowsType.SEQUENCTIAL])
    elif is_con:
        flows = _make_con(yaml_dict[FlowsType.CONCURRENT])
    elif is_onlyflow:
        flow_dict = {"flows": [yaml_dict]}
        flows = _make_seq(flow_dict)
    else:
        _logger.error("flows not defined")
        raise TypeError

    if "macros" in yaml_dict:
        mcr = Macros()
        mcr.update(yaml_dict["macros"])
        ctx_macros.set(mcr)

    return flows


def _make_flows(flows: dict) -> list[Flows | Flow]:
    flow_list = []
    for flow in flows:
        new_flow = None
        if "flow" in flow:
            new_flow = _make_flow(flow)
        elif FlowsType.SEQUENCTIAL in flow:
            new_flow = _make_seq(flow[FlowsType.SEQUENCTIAL])
        elif FlowsType.CONCURRENT in flow:
            new_flow = _make_con(flow[FlowsType.CONCURRENT])
        else:
            continue
        flow_list.append(new_flow)
    return flow_list


def _make_seq(config: dict) -> Flows:
    if "flows" not in config:
        _logger.error("no flows in sequential flow")
    flows = SequentialFlows()
    flow_list = _make_flows(config["flows"])
    flows.extend(flow_list)
    return flows


def _make_con(config: dict) -> Flows:
    if "flows" not in config:
        _logger.error("no flows in concurrent flow")
    flows = ConcurrentFlows()
    flow_list = _make_flows(config["flows"])
    flows.extend(flow_list)
    return flows


def _make_flow(config: dict) -> Flow:
    if "flow" not in config:
        _logger.error("no flow in flow")
    name = config.get("name", None)
    flow = Flow(name=name)
    prev_node = "root"
    for node in config["flow"]:
        if len(node) != 1:
            _logger.error("node read error")
        for node_type, params in node.items():
            parent = prev_node
            name = node_type
            if "name" in params:
                name = params["name"]
                del params["name"]
            if "parent" in params:
                parent = params["parent"]
                del params["parent"]
            flow.create_node(node_type, _parent=parent, name=name, param_dict=params)
            prev_node = name
    return flow


def _load_yaml(file_path: str | Path) -> dict:
    with open(file_path, "r") as file:
        data = yaml.safe_load(file)
    return data



from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from yaml import safe_load

from cauliflow.node import COMMON_ARGUMENT_DOC, ArgSpec, NodeFactory
from cauliflow.plugin_manager import PluginManager


class CauliFlowNodeDirective(SphinxDirective):
    required_arguments = 1
    has_content = False

    def run(self):
        node_type = self.arguments[0]
        node = NodeFactory.get(node_type)

        doc_src = None
        if node.__doc__ is not None:
            doc_src = safe_load(node.__doc__)

        parameters = {}

        common_arg_doc = safe_load(COMMON_ARGUMENT_DOC)
        parameters.update(common_arg_doc)

        doc_data = doc_src.get("DOCUMENTATION", None) if doc_src else None
        if doc_data:
            own_arg_doc = doc_data["parameters"] if "parameters" in doc_data else {}
            parameters.update(own_arg_doc)

        example = doc_src.get("EXAMPLE", None) if doc_src else None

        section = nodes.section()
        section["ids"].append(node_type)
        section += nodes.title(text=node_type)

        desc_txt = ""
        if doc_data is not None:
            desc_txt = doc_data.get("short_description", "No description")
        desc_param = nodes.paragraph(text=desc_txt)
        section += desc_param

        instance = node("instance")  # type: ignore
        instance.set_argument_spec()

        desc = doc_data.get("description", None) if doc_data is not None else None
        section += create_desc_section(desc)
        section += create_param_section(instance.argument_spec, parameters)
        section += create_examples_section(example)

        return [section]


def create_desc_section(desc: list[str] | None) -> nodes.section:
    section = nodes.section()
    section["ids"].append("description")
    section += nodes.title(text="Description")

    if desc is None:
        return section

    desc_bullet = nodes.bullet_list()
    for item_text in desc:
        item = nodes.list_item()
        para = nodes.paragraph(text=item_text)
        item += para
        desc_bullet += item

    section += desc_bullet

    return section


def create_param_section(args: dict[str, ArgSpec], doc_param: dict) -> nodes.section:
    colums = ["Parameter", "Type", "Required", "Default", "Comment"]
    no_colums = len(colums)

    # Table
    table = nodes.table()

    # Table group
    tgroup = nodes.tgroup(cols=no_colums)
    table += tgroup

    for _ in range(no_colums):
        tgroup += nodes.colspec(colwidth=1)

    # Header
    thead = nodes.thead()
    tgroup += thead

    header_row = nodes.row()
    for text in colums:
        entry = nodes.entry()
        entry += nodes.paragraph(text=text)
        header_row += entry
    thead += header_row

    # Body
    tbody = nodes.tbody()
    tgroup += tbody

    for name, data in sorted(args.items()):
        row = nodes.row()

        name_entry = nodes.entry()
        type_entry = nodes.entry()
        required_entry = nodes.entry()
        default_entry = nodes.entry()
        comment_entry = nodes.entry()

        required_txt = "Yes" if data.required else "No"
        default_txt = "-" if data.required else data.default
        comment_txt_list = (
            doc_param[name].get("description", "") if name in doc_param else ""
        )

        # name_entry += nodes.paragraph(text=str(name))
        name_entry += nodes.paragraph()
        name_entry += nodes.strong(text=str(name))
        type_entry += nodes.paragraph(text=data.type)
        required_entry += nodes.paragraph(text=str(required_txt))
        default_entry += nodes.paragraph(text=str(default_txt))
        for line in comment_txt_list:
            comment_entry += nodes.paragraph(text=line)

        row += name_entry
        row += type_entry
        row += required_entry
        row += default_entry
        row += comment_entry
        tbody += row

    section = nodes.section()
    section["ids"].append("parameters")
    section += nodes.title(text="Parameters")
    section += table

    return section


def create_examples_section(examples: str | None) -> nodes.section:
    section = nodes.section()
    section["ids"].append("examples")
    section += nodes.title(text="Examples")
    text = examples.strip() if examples else "No Examples"
    section += nodes.literal_block(text=text, language="yaml")
    return section


def setup(app):
    pm = PluginManager()
    pm.init()

    app.add_directive("cauliflow-node", CauliFlowNodeDirective)
    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from functools import singledispatch
from typing import cast

import janus
from aioca import CANothing, caget, camonitor, caput

from cauliflow.context import ctx_flowdata, init_flowdata
from cauliflow.node import ArgSpec, ProcessNode, TriggerNode, node


@node.register("camonitor")
class CamonitorNode(TriggerNode):
    """
    DOCUMENTATION:
      short_description: Monitor EPICS PV(s) and pass it flowdata.
      description:
        - Monitor EPICS PV(s) and pass it flowdata.
      parameters:
        pvname:
          description:
            - A pvname or a list of pvname.
    EXAMPLE: |-
      # Monitor single pv data.
      # Output: {'name': 'TEST:PV1', 'value': 7.0, 'timestamp': 1749196716.822903, 'status': 0, 'severity': 0, 'ok': True}
      - camonitor:
          pvname: "TEST:PV1"

      # Monitor list of pv data.
      # Output: {'name': 'TEST:PV1', 'value': 7.0, 'timestamp': 1749196716.822903, 'status': 0, 'severity': 0, 'ok': True}
      - camonitor:
          pvname: ["TEST:PV1", "TEST:PV2"]
    """

    def __init__(self, name: str, param_dict: dict | None = None):
        super().__init__(name, param_dict)
        self.q = janus.Queue()
        self.pvs = []

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "pvname": ArgSpec(type="any", required=True),
        }

    def callback(self, val, _):
        if val.ok:
            data = {
                "name": val.name,
                "value": val,
                "timestamp": val.timestamp,
                "status": val.status,
                "severity": val.severity,
                "ok": True,
            }
        else:
            data = {"name": val.name, "ok": False}

        self.q.sync_q.put(data)

    async def process(self):
        pvnames = self.params["pvname"]
        pvnames = _get_pvnames(pvnames)

        camonitor(
            pvnames, self.callback, format=1, all_updates=True, notify_disconnect=True
        )

        while True:
            pvdata = await self.q.async_q.get()
            init_flowdata()
            d = ctx_flowdata.get()
            d[self.name] = pvdata
            if self.child is None:
                continue
            await self.child.run()


@node.register("caget")
class CagetNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Get EPICS PV(s) and pass it flowdata.
      description:
        - Get EPICS PV(s) and pass it flowdata.
      parameters:
        pvname:
          description:
            - A pvname or a list of pvname.
        timeout:
          description:
            - Wait time.
    EXAMPLE: |-
      # Get single pv data.
      # Output: {'name': 'TEST:PV1', 'value': 7.0, 'timestamp': 1749196716.822903, 'status': 0, 'severity': 0, 'ok': True}
      - caget:
          pvname: "TEST:PV1"

      # Get list of pv data.
      # Output: [
      #  {'name': 'TEST:PV1', 'value': 7.0, 'timestamp': 1749196716.822903, 'status': 0, 'severity': 0, 'ok': True},
      #  {'name': 'TEST:PV2', 'value': 8.0, 'timestamp': 1749196717.822903, 'status': 0, 'severity': 0, 'ok': True},
      # ]
      - caget:
          pvname: ["TEST:PV1", "TEST:PV2"]

      # Get single pv data with timeout.
      # Timeout Output: {'name': 'TEST:PV1', 'ok': False}
      - caget:
          pvname: "TEST:PV1"
          timeout: 1.0
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "pvname": ArgSpec(type="str|list[str]", required=True),
            "timeout": ArgSpec(type="float", required=False, default=5.0),
        }

    async def process(self):
        pvnames = self.params["pvname"]
        is_single = isinstance(pvnames, str)
        pvnames = _get_pvnames(pvnames)

        vals = await caget(
            pvnames, format=1, timeout=self.params["timeout"], throw=False
        )

        out = []
        for val in vals:
            if val.ok:
                data = {
                    "name": val.name,
                    "value": val,
                    "timestamp": val.timestamp,
                    "status": val.status,
                    "severity": val.severity,
                    "ok": True,
                }
            else:
                data = {"name": val.name, "ok": False}
            out.append(data)

        if is_single:
            out = out[0]

        fd = ctx_flowdata.get()
        fd[self.name] = out


@node.register("caput")
class CaputNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Put values to EPICS PV(s).
      description:
        - Put values to EPICS PV(s).
      parameters:
        pvname:
          description:
            - A pvname or a list of pvname.
        value:
          description:
            - A value or a list of value.
        repeat_value:
          description:
            - Put the same value to every PV if True.
        timeout:
          description:
            - Wait time.
    EXAMPLE: |-
      # Put to a single PV.
      # Output: {'name': 'TEST:PV1', 'ok': True}
      - caput:
          pvname: "TEST:PV1"
          value: 1.0

      # Put 1 to TEST:PV1 and 2 to TEST:PV2.
      # Output: [
      #  {'name': 'TEST:PV1', 'ok': True}
      #  {'name': 'TEST:PV2', 'ok': True}
      # ]
      - caput:
          pvname: ["TEST:PV1", "TEST:PV2"]
          value: [1.0, 2.0]

      # Put 1 to TEST:PV1 and TEST:PV2 with repeat_value.
      # Output: [
      #  {'name': 'TEST:PV1', 'ok': True}
      #  {'name': 'TEST:PV2', 'ok': True}
      # ]
      - caput:
          pvname: ["TEST:PV1", "TEST:PV2"]
          value: 1.0
          repeat_value: yes

      # Put to a single pv data with timeout.
      # Timeout Output: {'name': 'TEST:PV1', 'ok': False}
      - caput:
          pvname: "TEST:PV1"
          value: 1.0
          timeout: 1.0
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "pvname": ArgSpec(type="str|list[str]", required=True),
            "value": ArgSpec(type="Any|list[Any]", required=True),
            "repeat_value": ArgSpec(type="bool", required=False),
            "timeout": ArgSpec(type="float", required=False, default=5.0),
        }

    async def process(self):
        pvnames = self.params["pvname"]
        is_single = isinstance(pvnames, str)

        vals = None
        if is_single:
            vals = await caput(
                pvnames,
                self.params["value"],
                wait=True,
                timeout=self.params["timeout"],
                throw=False,
            )
        else:
            vals = await caput(
                pvnames,
                self.params["value"],
                repeat_value=self.params["repeat_value"],
                wait=True,
                timeout=self.params["timeout"],
                throw=False,
            )

        out = None
        if is_single:
            vals = cast(CANothing, vals)
            out = {"name": vals.name, "ok": vals.ok}
        else:
            vals = cast(list[CANothing], vals)
            out = [{"name": val.name, "ok": val.ok} for val in vals]

        fd = ctx_flowdata.get()
        fd[self.name] = out


@singledispatch
def _get_pvnames(obj):
    return obj


@_get_pvnames.register(str)
def _(string):
    return [string]



from functools import singledispatchmethod

from zabbix_utils import AsyncSender, AsyncZabbixAPI, ItemValue

from cauliflow.logging import get_logger
from cauliflow.node import ArgSpec, ProcessNode, node

_logger = get_logger(__name__)


@node.register("zabbix_get_item")
class ZabbixGetItemNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Get Zabbix item with Zabbix API
      description:
        - Get Zabbix item with Zabbix API
      parameters:
        url:
          description:
            - URL for Zabbix server
        user:
          description:
            - User name for Zabbix API.
        password:
          description:
            - Password for Zabbix API.
        filter:
          description:
            - Filter to get the result mathed.
        output:
          description:
            - Properties to be returned.
    EXAMPLE: |-
      # Get Zabbix items only matched Template EPICS
      # Output: [{'key_': 'item.key1',  'name': 'foo, 'itemid': 1}]
      - zabbix_get_item:
          name: "zabbix_get"
          url: "http:/localhost"
          user: "Admin"
          password: "Zabbix"
          output: ["itemid", "name", "key_"]
          filter: { "key_": null, "host": "Template EPICS" }
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {
            "url": ArgSpec(type="str", required=False, default="localhost"),
            "user": ArgSpec(type="str", required=False, default="root"),
            "password": ArgSpec(type="str", required=False, default="Zabbix"),
            "filter": ArgSpec(type="dict", required=False, default=None),
            "output": ArgSpec(type="list[str]", required=False, default=None),
        }

    async def process(self) -> None:
        self.api = AsyncZabbixAPI(url=self.params["url"])
        await self.api.login(user=self.params["user"], password=self.params["password"])

        items = await self.api.item.get(  # type: ignore
            output=self.params["output"], filter=self.params["filter"]
        )

        await self.api.logout()
        self.output(items)


@node.register("zabbix_send")
class ZabbixSend(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Send item values to a Zabbix server.
      description:
        - Send item values to a Zabbix server via Zabbix sender protocol.
      parameters:
        server:
          description:
            - Zabbix server address.
        port:
          description:
            - Zabbix server port.
        items:
          description:
            - List of items or dict of item to send.
            - "Item must have following keys: hostname, key, and value."
    EXAMPLE: |-
      # Send two items to Zabbix server.
      # Output: No output
      - zabbix_send:
          name: "zabbix_send"
          server: "localhost"
          port: 10051
          items:
            - {"hostname": "foo", "key": "bar", "val": 1}
            - {"hostname": "foo", "key": "foobar", "val": 1}

      # Send items from flowdata to Zabbix server.
      # Output: No output
      - zabbix_send:
          name: "zabbix_send"
          items: "{{ fd.zabbix_item }}"
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "server": ArgSpec(type="str", required=False, default="localhost"),
            "port": ArgSpec(type="int", required=False, default=10051),
            "items": ArgSpec(type="dict|list[dict]", required=True),
        }

    async def process(self) -> None:
        sender = AsyncSender(server=self.params["server"], port=self.params["port"])

        input = self.params["items"]
        items = self._create_items(input)
        response = await sender.send(items)
        _logger.debug(response)

    @singledispatchmethod
    def _create_items(self, item: dict) -> list[ItemValue]:
        return [ItemValue(item["hostname"], item["key"], item["value"])]

    @_create_items.register
    def _(self, items: list) -> list[ItemValue]:
        item_list = []
        for item in items:
            item_list.append(ItemValue(item["hostname"], item["key"], item["value"]))

        return item_list



from copy import deepcopy
from functools import singledispatchmethod

from cauliflow.node import ArgSpec, ProcessNode, node


@node.register("concat")
class ConcatNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Concatenate strings.
      description:
        - Concatenate strings.
        - list of string can be passed.
      parameters:
        first:
          description:
            - string or list of strings to be concatnate before second parameter
        second:
          description:
            - string or list of strings to be concatnate after first parameter
    EXAMPLE: |-
      # Concatenate the 2 strings.
      # Output: {'concat': 'head1:pv1'}
      - concat:
          name: "concat"
          first: "head1:"
          second: "pv1"

      # Concatenate the 1 string and 1 list of string.
      # Output: {'concat': ['head1:pv1', 'head1:pv2']}
      - concat:
          name: "concat"
          first: "head1:"
          second: ["pv1", "pv2"]

      # Concatenate the 2 list of string.
      # Output: {'concat': ['head1:pv1', 'head2:pv2']}
      - concat:
          name: "concat"
          first: ["head1:", "head2:"]
          second: ["pv1", "pv2"]

      # Concatenate the 2 list of string, but length is different.
      # Output: {'concat': ['head1:pv1', 'head2:pv2']}
      - concat:
          name: "concat"
          first: ["head1:", "head2:", "head3:"]
          second: ["pv1", "pv2"]

      # Concatenate the 2 strings and output to the blackboard.
      # Output: {'foo': 'head1:pv1'}
      - concat:
          name: "concat"
          first: "head1:"
          second: "pv1"
          out_bb: yes
          out_field: "foo"
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {
            "first": ArgSpec(type="str|list[str]", required=True),
            "second": ArgSpec(type="str|list[str]", required=True),
        }

    async def process(self) -> None:
        first = self.params["first"]
        second = self.params["second"]
        out = self._concat(first, second)
        self.output(out)

    def _concat(
        self, first: str | list[str], second: str | list[str]
    ) -> str | list[str]:
        is_1st_list = isinstance(first, list)
        is_2nd_list = isinstance(second, list)

        if is_1st_list and is_2nd_list:
            min_len = min(len(first), len(second))
            out = []
            for i in range(min_len):
                out.append(first[i] + second[i])
            return out

        if is_1st_list:
            return [fl + second for fl in first]  # type: ignore

        if is_2nd_list:
            return [first + sl for sl in second]  # type: ignore

        return first + second


@node.register("mutate")
class MutateNode(ProcessNode):
    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {
            "target": ArgSpec(type="any", required=True),
            "split": ArgSpec(type="dict", required=False, default={}),
            "copy": ArgSpec(type="dict", required=False, default={}),
        }

    async def process(self) -> None:
        target = self.params["target"]
        target = deepcopy(target)
        self.apply(target)
        self.output(target)

    @singledispatchmethod
    def apply(self, target: dict):
        self.copy(target)
        self.split(target)

    @apply.register
    def _(self, targets: list) -> None:
        for target in targets:
            self.apply(target)

    def split(self, target) -> None:
        split_dict = self.params["split"]
        for field, parser in split_dict.items():
            target[field] = target[field].split(parser)

    def copy(self, target):
        copy_dict = self.params["copy"]
        for src, dst in copy_dict.items():
            target[dst] = deepcopy(target[src])



import asyncio
from collections.abc import Iterable
from enum import StrEnum
from typing import TypedDict

from cauliflow.context import ctx_flowdata, init_flowdata
from cauliflow.flowdata import FlowData
from cauliflow.logging import get_logger
from cauliflow.node import ArgSpec, FlowControlNode, Node, node
from cauliflow.variable import Variable

_logger = get_logger(__name__)


class TypeChild(TypedDict):
    child_if: Node | None
    child_else: Node | None


@node.register("if")
class IfNode(FlowControlNode):
    """
    DOCUMENTATION:
      short_description: Control flow with if-then-else construction.
      description:
        - Control flow with if-then-else construction.
        - Run nodes of children followed child_if or child_else.
        - When the flow of child_if or child_else is finished, then run children of if node.
      parameters:
        condition:
          description:
            - The condition that determines the execution path in a conditional branch.
        child_if:
          description:
            - Reference to the node to be processed if the condition parameter evaluates to True.
        child_else:
          description:
            - Reference to the node to be processed if the condition parameter evaluates to False.
    EXAMPLE: |-
      # Print true if flowdata['cond'] is OK else print false.
      # Parent node can be set with parent field as <node name>.<field name>
      #     ┌─> true message  ─┐
      # if ─┤                  ├─> out
      #     └─> false message ─┘
      # Output: No output
      - if:
          name: "if"
          condition: "fd.cond == 'OK'"
      - message:
          name: "true message"
          msg: "true"
          out_field: "message"
          parent: "if.child_if"
      - message:
          name: "false message"
          msg: "false"
          out_field: "message"
          parent: "if.child_else"
      - stdout:
          name: "out"
          src: "{{ fd.message }}"
          parent: "if"

      # This example behaves in the same way as the one above.
      #     ┌─> true message  ──> out true
      # if ─┤
      #     └─> false message ──> out false
      - if:
          name: "if"
          condition: "macro.cond == 'OK'"
      - message:
          name: "true message"
          msg: "true"
          parent: "if.child_if"
      - stdout:
          name: "out true"
          src: "{{ fd['true message'] }}"
      - message:
          name: "false message"
          msg: "false"
          parent: "if.child_else"
      - stdout:
          name: "out false"
          src: "{{ fd['false message'] }}"
    """

    def __init__(self, name: str, param_dict: dict | None = None):
        super().__init__(name, param_dict)
        self.cond: Variable | None = None
        self.if_children: TypeChild = {"child_if": None, "child_else": None}

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "condition": ArgSpec(type="expression", required=True),
            "child_if": ArgSpec(type="node", required=False, default=None),
            "child_else": ArgSpec(type="node", required=False, default=None),
        }

    async def process(self) -> None:
        if self.cond is None:
            self.cond = Variable("{{" + self.params["condition"] + "}}")

        cond = self.cond.fetch()

        if cond and self.if_children["child_if"] is not None:
            await self.if_children["child_if"].run()
        elif self.if_children["child_else"] is not None:
            await self.if_children["child_else"].run()

    def add_child(self, child: Node, param: str | None = None) -> None:
        if param is None:
            self.child = child
            return

        if param not in self.if_children.keys():
            _logger.warning(f"{param} is invalid field")

        self.if_children[param] = child


class ForEachMode(StrEnum):
    SEQUENTIAL = "sequential"
    CONCURRENT = "concurrent"


@node.register("foreach")
class ForEachNode(FlowControlNode):
    """
    DOCUMENTATION:
      short_description: Run the child node for each item in the list of data.
      description:
        - Run the child node for each item in the list of data.
        - The child node can be called either sequentially or concurrently.
        - The item in the list is passed within the flowdata.
        - When the flow of child_for is finished, then run children of foreach node.
      parameters:
        mode:
          description:
            - Set either of sequential or concurrent.
        items:
          description:
            - List of elements to be processed one by one by the child node.
        item_name:
          description:
            - Key name of passed item in the flowdata.
        child_for:
          description:
            - Child node to be run for each item.
    EXAMPLE: |-
      # Print 1, 2, 3, and hello into the stdout
      # Output: No output
      - foreach:
          name: "for"
          items: [1, 2, 3]
          item_name: "foo"
          mode: "sequential"
      - stdout:
          name: "out_for"
          src: "{{ fd.foo }}"
          parent: "for.child_for"
      - stdout:
          name: "out"
          src: "hello"
          parent: "for"
    """

    def __init__(self, name: str, param_dict: dict | None = None):
        super().__init__(name, param_dict)
        self.child_for: Node | None = None

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "mode": ArgSpec(type="str", required=False, default=ForEachMode.CONCURRENT),
            "items": ArgSpec(type="list", required=True),
            "item_name": ArgSpec(type="str", required=False, default="item"),
            "child_for": ArgSpec(type="node", required=False, default=None),
        }

    async def process(self) -> None:
        if self.child_for is None:
            return

        items = self.params["items"]
        if not isinstance(items, Iterable):
            raise ValueError("items is not Iterable")

        mode = self.params["mode"]
        item_name = self.params["item_name"]

        base_fd = ctx_flowdata.get()

        if mode == ForEachMode.SEQUENTIAL:
            await self._sequential(base_fd, items, item_name, self.child_for)
        elif mode == ForEachMode.CONCURRENT:
            await self._concurrent(base_fd, items, item_name, self.child_for)
        else:
            raise ValueError(f"{mode} is not valid mode")

        ctx_flowdata.set(base_fd)

    def add_child(self, child: Node, param: str | None = None) -> None:
        if param is None:
            self.child = child
            return

        if param != "child_for":
            _logger.warning(f"{param} is invalid field")

        self.child_for = child

    async def _sequential(self, base_fd, items, item_name, child: Node):
        for item in items:
            init_flowdata()
            fd = ctx_flowdata.get()
            fd.update(base_fd)
            fd[item_name] = item
            await child.run()

    async def _concurrent(self, base_fd, items, item_name, child: Node):
        async with asyncio.TaskGroup() as tg:
            for item in items:
                init_flowdata()
                fd = ctx_flowdata.get()
                fd.update(base_fd)
                fd[item_name] = item
                tg.create_task(child.run())


@node.register("dispatch")
class DispatchNode(FlowControlNode):
    """
    DOCUMENTATION:
      short_description: Dispatch the flowdata to the target nodes.
      description:
        - Dispatch the flowdata to the flows of target nodes.
        - The target nodes can be called either sequentially or concurrently.
        - The flowdata is shared in sequential mode.
        - The flowdata is finally merged and passed to the child node in concurrent mode.
        - When the flows of the target nodes are finished, then run children of dispatch node.
      parameters:
        mode:
          description:
            - Set either of sequential or concurrent.
        targets:
          description:
            - Target nodes to dispatch the flowdata.
    EXAMPLE: |-
      # Dispatch the flowdata sequentially
      # Output: {'msg': 10, 'target1': 11, 'target2': 21, 'target3': 31}
      - message:
          name: "msg"
          msg: 10
      - dispatch:
          name: "dispatch"
          mode: "sequential"
      - message:
          name: "target1"
          msg: "{{ fd.msg + 1 }}"
          parent: "dispatch.targets"
      - message:
          name: "target2"
          msg: "{{ fd.msg + fd.target1 }}"
          parent: "dispatch.targets"
      - message:
          name: "target3"
          msg: "{{ fd.msg + fd.target2 }}"
          parent: "dispatch.targets"

      # Dispatch the flowdata concurrently
      # Output: {'msg': 10, 'target1': 11, 'target2': 12, 'target3': 13}
      - message:
          name: "msg"
          msg: 10
      - dispatch:
          name: "dispatch"
          mode: "concurrent"
      - message:
          name: "target1"
          msg: "{{ fd.msg + 1 }}"
          parent: "dispatch.targets"
      - message:
          name: "target2"
          msg: "{{ fd.msg + 2 }}"
          parent: "dispatch.targets"
      - message:
          name: "target3"
          msg: "{{ fd.msg + 3 }}"
          parent: "dispatch.targets"
    """

    def __init__(self, name: str, param_dict: dict | None = None):
        super().__init__(name, param_dict)
        self.targets: list[Node] = []

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "mode": ArgSpec(type="str", required=False, default=ForEachMode.CONCURRENT),
            "targets": ArgSpec(type="list(node)", required=False, default=None),
        }

    async def process(self) -> None:
        if len(self.targets) < 1:
            _logger.warning("dispatch is called, but it has no target.")
            return

        mode = self.params["mode"]

        if mode == ForEachMode.SEQUENTIAL:
            await self._sequential(self.targets)
        elif mode == ForEachMode.CONCURRENT:
            base_fd = ctx_flowdata.get()
            fd = await self._concurrent(base_fd, self.targets)
            ctx_flowdata.set(fd)
        else:
            raise ValueError(f"{mode} is not valid mode")

    def add_child(self, child: Node, param: str | None = None) -> None:
        if param is None:
            self.child = child
            return

        if param != "targets":
            _logger.warning(f"{param} is invalid field")

        self.targets.append(child)

    async def _sequential(self, targets: list[Node]) -> None:
        for target in targets:
            await target.run()

    async def _concurrent(self, base_fd, targets: list[Node]) -> FlowData:
        fds = []
        async with asyncio.TaskGroup() as tg:
            for target in targets:
                init_flowdata()
                fd = ctx_flowdata.get()
                fds.append(fd)
                fd.update(base_fd)
                tg.create_task(target.run())
        ret = {}
        for fd in fds:
            ret.update(fd)
        return FlowData(ret)



import pytest

from cauliflow.context import ctx_blackboard, ctx_flowdata
from cauliflow.node import ArgSpec, Node


class NodeTest(Node):
    async def process(self) -> None:
        pass

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {
            "required_str": ArgSpec(type="str", required=True),
            "not_required_str": ArgSpec(type="str", required=False, default="default"),
            "not_required_bool": ArgSpec(type="bool", required=False, default=False),
            "not_required_float": ArgSpec(type="float", required=False, default=1.1),
        }


@pytest.mark.asyncio
async def test_node_with_positive_args():
    args = {
        "required_str": "foo",
        "not_required_str": "bar",
        "not_required_bool": True,
        "not_required_float": 2.2,
    }
    node = NodeTest(name="msg", param_dict=args)
    await node.run()
    assert node.params["required_str"] == "foo"
    assert node.params["not_required_str"] == "bar"
    assert node.params["not_required_bool"] is True
    assert node.params["not_required_float"] == 2.2


@pytest.mark.asyncio
async def test_node_without_default():
    args = {
        "required_str": "foo",
    }
    node = NodeTest(name="msg", param_dict=args)
    await node.run()
    assert node.params["required_str"] == "foo"
    assert node.params["not_required_str"] == "default"
    assert node.params["not_required_bool"] is False
    assert node.params["not_required_float"] == 1.1


@pytest.mark.asyncio
async def test_node_without_required():
    with pytest.raises(TypeError) as err:
        args = {}
        node = NodeTest(name="msg", param_dict=args)
        await node.run()
    assert "missing parameter required by" in str(err.value)


@pytest.mark.parametrize(
    "out_bb, out_field, ctx, expected_field",
    [
        (False, None, ctx_flowdata, "msg"),
        (False, "field", ctx_flowdata, "field"),
        (True, None, ctx_blackboard, "msg"),
        (True, "field", ctx_blackboard, "field"),
    ],
)
def test_output_bb(out_bb, out_field, ctx, expected_field, init_context_vars):
    args = {"required_str": "foo", "out_bb": out_bb, "out_field": out_field}
    node = NodeTest(name="msg", param_dict=args)
    node._fetch_params()
    node.output("test")
    dest = ctx.get()
    assert dest[expected_field] == "test"



import pytest

from cauliflow.context import ctx_blackboard, ctx_flowdata
from cauliflow.flow import Flow


@pytest.mark.asyncio
async def test_if_child_if(init_plugins):
    flow = Flow("test")

    args1 = {"a": 1, "b": 1}
    args2 = {"condition": "fd['add1']  == 2"}
    args3 = {"a": "{{ fd['add1'] }}", "b": 1}
    args4 = {"a": "{{ fd['add1'] }}", "b": 2}
    args5 = {"a": 1, "b": 1}

    flow.create_node("test.addnode", "root", "add1", args1)
    flow.create_node("if", "add1", "if", args2)
    flow.create_node("test.addnode", "if.child_if", "add3", args3)
    flow.create_node("test.addnode", "if.child_else", "add4", args4)
    flow.create_node("test.addnode", "if", "add5", args5)

    await flow.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["add1"] == 2
    assert flowdata["add3"] == 3
    assert "add4" not in flowdata
    assert flowdata["add5"] == 2


@pytest.mark.asyncio
async def test_if_child_else(init_plugins):
    flow = Flow("test")

    args1 = {"a": 1, "b": 1}
    args2 = {"condition": "fd['add1']  != 2"}
    args3 = {"a": "{{ fd['add1'] }}", "b": 1}
    args4 = {"a": "{{ fd['add1'] }}", "b": 2}
    args5 = {"a": 1, "b": 1}

    flow.create_node("test.addnode", "root", "add1", args1)
    flow.create_node("if", "add1", "if", args2)
    flow.create_node("test.addnode", "if.child_if", "add3", args3)
    flow.create_node("test.addnode", "if.child_else", "add4", args4)
    flow.create_node("test.addnode", "if", "add5", args5)

    await flow.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["add1"] == 2
    assert "add3" not in flowdata
    assert flowdata["add4"] == 4
    assert flowdata["add5"] == 2


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "mode",
    [
        ("sequential"),
        ("concurrent"),
    ],
)
async def test_foreach(init_plugins, mode):
    flow = Flow("test")

    args_for = {"items": [1, 2, 3], "item_name": "foo", "mode": mode}
    args_msg = {
        "msg": "{{ fd.foo }}",
        "out_bb": True,
        "out_field": "{{ 'field' + fd.foo | str }}",
    }

    flow.create_node("foreach", "root", "for", args_for)
    flow.create_node("message", "for.child_for", "msg", args_msg)

    await flow.run()

    bb = ctx_blackboard.get()
    assert bb == {"field1": 1, "field2": 2, "field3": 3}
    fd = ctx_flowdata.get()
    assert fd == {}


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "items, mode",
    [
        ([1, 2, 3], "invalid_mode"),
        (1, "concurrent"),
    ],
)
async def test_foreach_valueerror(init_plugins, items, mode):
    flow = Flow("test")

    args_for = {"items": items, "item_name": "foo", "mode": mode}
    args_msg = {
        "msg": "{{ fd.foo }}",
        "out_bb": True,
        "out_field": "{{ 'field' + fd.foo | str }}",
    }

    flow.create_node("foreach", "root", "for", args_for)
    flow.create_node("message", "for.child_for", "msg", args_msg)

    with pytest.raises(ValueError):
        await flow.run()


@pytest.mark.asyncio
async def test_dispatch_sequential(init_plugins):
    flow = Flow("test")

    args_msg = {"msg": 10}
    args_dispatch = {"mode": "sequential"}
    args1 = {"a": "{{ fd.msg }}", "b": 1}
    args2 = {"a": "{{ fd.msg }}", "b": "{{ fd.add1 }}"}
    args3 = {"a": "{{ fd.msg }}", "b": "{{ fd.add2 }}"}

    flow.create_node("message", "root", "msg", args_msg)
    flow.create_node("dispatch", "msg", "dispatch", args_dispatch)
    flow.create_node("test.addnode", "dispatch.targets", "add1", args1)
    flow.create_node("test.addnode", "dispatch.targets", "add2", args2)
    flow.create_node("test.addnode", "dispatch.targets", "add3", args3)

    await flow.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["add1"] == 11
    assert flowdata["add2"] == 21
    assert flowdata["add3"] == 31


@pytest.mark.asyncio
async def test_dispatch_concurrent(init_plugins):
    flow = Flow("test")

    args_msg = {"msg": 10}
    args_dispatch = {"mode": "concurrent"}
    args1 = {"a": "{{ fd.msg }}", "b": 1}
    args2 = {"a": "{{ fd.msg }}", "b": 2}
    args3 = {"a": "{{ fd.msg }}", "b": 3}

    flow.create_node("message", "root", "msg", args_msg)
    flow.create_node("dispatch", "msg", "dispatch", args_dispatch)
    flow.create_node("test.addnode", "dispatch.targets", "add1", args1)
    flow.create_node("test.addnode", "dispatch.targets", "add2", args2)
    flow.create_node("test.addnode", "dispatch.targets", "add3", args3)

    await flow.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["add1"] == 11
    assert flowdata["add2"] == 12
    assert flowdata["add3"] == 13



import asyncio
import json

import pytest
from aioresponses import CallbackResult, aioresponses

from cauliflow.context import ctx_flowdata
from cauliflow.plugins.http import HTTPNode


@pytest.fixture
def mock_aioresponse():
    with aioresponses() as m:
        yield m


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "format, expected",
    [
        ("json", {"foo": "bar"}),
        ("text", '{"foo": "bar"}'),
    ],
)
async def test_get_data(init_context_vars, mock_aioresponse, format, expected):
    url = "http://example.com/api/data"
    expected_data = {"foo": "bar"}
    mock_aioresponse.get(url, payload=expected_data)

    node = HTTPNode(name="node", param_dict={"url": url, "format": format})
    await node.run()
    flowdata = ctx_flowdata.get()
    data = flowdata["node"]
    assert data["status"] == 200
    assert data["data"] == expected


@pytest.mark.asyncio
async def test_get_data_timeout(init_context_vars, mock_aioresponse):
    url = "http://example.com/api/data"
    mock_aioresponse.get(url, exception=asyncio.TimeoutError())

    node = HTTPNode(name="node", param_dict={"url": url, "format": format})
    await node.run()
    flowdata = ctx_flowdata.get()
    data = flowdata["node"]
    assert data["error"] == "TimeoutError"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "format, expected",
    [
        ("json", {"foo": "bar"}),
        ("text", '{"foo": "bar"}'),
    ],
)
async def test_put_data(init_context_vars, mock_aioresponse, format, expected):
    url = "http://example.com/api/data"
    expected_data = {"foo": "bar"}

    def request_callback(url, **kwargs):
        captured_body = kwargs.get("data", {})
        assert captured_body == "test"
        return CallbackResult(status=201, body=json.dumps(expected_data))

    mock_aioresponse.put(url, callback=request_callback)

    node = HTTPNode(
        name="node",
        param_dict={"url": url, "format": format, "method": "put", "body": "test"},
    )

    await node.run()
    flowdata = ctx_flowdata.get()
    data = flowdata["node"]
    assert data["status"] == 201
    assert data["data"] == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "format, expected",
    [
        ("json", {"foo": "bar"}),
        ("text", '{"foo": "bar"}'),
    ],
)
async def test_post_data(init_context_vars, mock_aioresponse, format, expected):
    url = "http://example.com/api/data"
    expected_data = {"foo": "bar"}

    def request_callback(url, **kwargs):
        captured_body = kwargs.get("data", {})
        assert captured_body == "test"
        return CallbackResult(status=201, body=json.dumps(expected_data))

    mock_aioresponse.post(url, callback=request_callback)

    node = HTTPNode(
        name="node",
        param_dict={"url": url, "format": format, "method": "post", "body": "test"},
    )

    await node.run()
    flowdata = ctx_flowdata.get()
    data = flowdata["node"]
    assert data["status"] == 201
    assert data["data"] == expected


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "format, expected",
    [
        ("json", {"foo": "bar"}),
        ("text", '{"foo": "bar"}'),
    ],
)
async def test_patch_data(init_context_vars, mock_aioresponse, format, expected):
    url = "http://example.com/api/data"
    expected_data = {"foo": "bar"}

    def request_callback(url, **kwargs):
        captured_body = kwargs.get("data", {})
        assert captured_body == "test"
        return CallbackResult(status=201, body=json.dumps(expected_data))

    mock_aioresponse.patch(url, callback=request_callback)

    node = HTTPNode(
        name="node",
        param_dict={"url": url, "format": format, "method": "patch", "body": "test"},
    )

    await node.run()
    flowdata = ctx_flowdata.get()
    data = flowdata["node"]
    assert data["status"] == 201
    assert data["data"] == expected


@pytest.mark.asyncio
async def test_delete_data(init_context_vars, mock_aioresponse):
    url = "http://example.com/api/data"
    mock_aioresponse.delete(url, status=204)

    node = HTTPNode(name="node", param_dict={"url": url, "method": "delete"})
    await node.run()
    flowdata = ctx_flowdata.get()
    data = flowdata["node"]
    assert data["status"] == 204



from os import makedirs
from pathlib import Path

from cauliflow.node import FlowControlNode, NodeFactory, ProcessNode, TriggerNode
from cauliflow.plugin_manager import PluginManager


def main():
    pm = PluginManager()
    pm.init()

    registry = NodeFactory.registry
    dir = Path("nodes")

    makedirs(dir, exist_ok=True)
    for klass_dir in ["trigger", "process", "control"]:
        makedirs(dir / klass_dir, exist_ok=True)

    for name, klass in registry.items():
        if name == "root":
            continue

        klass_dir = None
        if issubclass(klass, TriggerNode):
            klass_dir = "trigger"
        elif issubclass(klass, ProcessNode):
            klass_dir = "process"
        elif issubclass(klass, FlowControlNode):
            klass_dir = "control"
        else:
            raise KeyError

        p = dir / klass_dir / f"{name}.rst"
        with open(p, mode="w") as f:
            f.write(f".. cauliflow-node:: {name}")


if __name__ == "__main__":
    main()



import asyncio
from logging import DEBUG, WARNING, getLogger

import click

from cauliflow.context import ContextFlows, ctx_flows, ctx_macros
from cauliflow.loader import flow_from_yaml
from cauliflow.logging import get_logger
from cauliflow.plugin_manager import PluginManager

_logger = get_logger(__name__)


def init_logger(debug: bool = False) -> None:
    logger = getLogger("cauliflow")
    level = DEBUG if debug else WARNING
    logger.setLevel(level)


@click.group()
@click.version_option(package_name="cauliflow", message="%(prog)s %(version)s")
def cli():
    pass


@cli.command()
@click.option("--macro", "-m", "macros", type=(str, str), multiple=True)
@click.option("--debug/--no-debug", default=False)
@click.argument("filename", type=click.Path(exists=True))
def run(macros, debug, filename):
    init_logger(debug)
    pm = PluginManager()
    pm.init()

    ctx_flows.set(ContextFlows(debug=debug))
    flows = flow_from_yaml(filename)

    macros_dict = dict(macros)
    mcr = ctx_macros.get()
    mcr.update(macros_dict)
    ctx_macros.set(mcr)
    if debug:
        _logger.debug(f"macros={mcr}")
    asyncio.run(flows.run())


if __name__ == "__main__":
    cli()



import pytest

from cauliflow.context import ctx_flowdata
from cauliflow.plugins.transforms import (
    ConcatNode,
    MutateNode,
)


@pytest.mark.asyncio
async def test_mutate(init_context_vars):
    params = {
        "target": [{"origin": "head.pvname1"}, {"origin": "head.pvname2"}],
        "copy": {"origin": "copy"},
        "split": {"copy": "."},
    }
    node = MutateNode(name="node", param_dict=params)
    await node.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["node"] == [
        {"origin": "head.pvname1", "copy": ["head", "pvname1"]},
        {"origin": "head.pvname2", "copy": ["head", "pvname2"]},
    ]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "params, expected",
    [
        (
            {"first": ["head1:", "head2:"], "second": ["pv1", "pv2"]},
            ["head1:pv1", "head2:pv2"],
        ),
        (
            {"first": "head1:", "second": ["pv1", "pv2"]},
            ["head1:pv1", "head1:pv2"],
        ),
        (
            {"first": ["head1:", "head2:"], "second": "pv1"},
            ["head1:pv1", "head2:pv1"],
        ),
        (
            {"first": "head1:", "second": "pv1"},
            "head1:pv1",
        ),
    ],
)
async def test_concat(init_context_vars, params, expected):
    node = ConcatNode(name="node", param_dict=params)
    await node.run()
    flowdata = ctx_flowdata.get()
    assert flowdata["node"] == expected



import asyncio

from cauliflow.context import ctx_flowdata
from cauliflow.flowdata import FlowData
from cauliflow.node import ArgSpec, FlowControlNode, node


@node.register("buffer")
class BufferNode(FlowControlNode):
    """
    DOCUMENTATION:
      short_description: Buffering input data.
      description:
          - Buffering input data.
          - The child nodes are processed either when the data exceeds the buffer size or after a certain amount of time has passed. At that point, the buffer data is written to the FlowData.
      parameters:
        size:
          description:
            - Buffer size.
        timeout:
          description:
            - The waiting period for new data to arrive.
            - If this period expires, the child nodes are processed.
            - If the input data exceeds the buffer size, the child nodes are processed immediately.
        input:
          description:
            - Input data to be buffered.
        flatten:
          description:
            - When flatten is True, the input list is flattened before being stored in the buffer.
    EXAMPLE: |-
      # Buffer for scalar data
      - buffer:
          name: "buffer"
          size: 10
          timeout: 5
          input: "{{ fd.input }}"

      # Buffer for list data with flatten
      - buffer:
          name: "buffer"
          size: 10
          timeout: 5
          input: "{{ fd.input }}"
          flatten: yes
    """

    def __init__(self, name: str, param_dict: dict | None = None):
        super().__init__(name, param_dict)
        self.buffer = []
        self.lock = asyncio.Lock()
        self.timer_task = None

    # redeclare run not to run a child node
    async def run(self) -> None:
        await self._run_self()

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "size": ArgSpec(type="int", required=False, default=50),
            "timeout": ArgSpec(type="float", required=False, default=5.0),
            "input": ArgSpec(type="Any", required=True),
            "flatten": ArgSpec(type="bool", required=False, default=False),
        }

    async def process(self) -> None:
        input = self.params["input"]
        size = self.params["size"]
        timeout = self.params["timeout"]
        flatten = self.params["flatten"]

        async with self.lock:
            if flatten:
                self.buffer.extend(input)
            else:
                self.buffer.append(input)

            if len(self.buffer) >= size:
                await self.flush()
            elif self.timer_task is None:
                self.timer_task = asyncio.create_task(self.start_timer(timeout))

    async def start_timer(self, timeout):
        await asyncio.sleep(timeout)
        async with self.lock:
            if self.buffer:
                await self.flush()

    async def flush(self):
        if self.buffer:
            # buffer node resets the flowdata with buffer data
            flowdata = {self.name: self.buffer}
            ctx_flowdata.set(FlowData(flowdata))
            self.buffer = []

            if self.timer_task:
                self.timer_task.cancel()
            self.timer_task = None

            await self._run_child()



from collections.abc import Callable

from docutils import nodes
from sphinx.util.docutils import SphinxDirective
from yaml import safe_load

from cauliflow.filters import FILTERS


class CauliFlowFilterDirective(SphinxDirective):
    has_content = False

    def run(self):
        section = nodes.section()
        section["ids"].append("filters")
        section += nodes.title(text="Filters")
        for name, func in FILTERS.items():
            sec = create_filter_sec(name, func)
            section += sec
        return [section]


def create_filter_sec(name: str, func: Callable) -> nodes.section:
    section = nodes.section()
    section["ids"].append(name)
    section += nodes.title(text=name)

    doc = None
    if func.__doc__ is not None:
        doc = safe_load(func.__doc__)

    if doc is None:
        return section

    desc = doc.get("description", None)

    if desc:
        for item_text in desc:
            para = nodes.paragraph(text=item_text)
            section += para

    params = create_params(doc)
    if params is not None:
        section += params

    example = create_example(doc)
    if example is not None:
        section += example

    return section


def create_params(doc: dict) -> nodes.definition_list | None:
    params = doc.get("parameters", None)
    if params is None:
        return None

    bullet_list = nodes.bullet_list()
    for name, info in params.items():
        para = nodes.paragraph()
        para += nodes.strong(text=name)
        para += nodes.Text(f" – {info['description']}")

        item = nodes.list_item()
        item += para
        bullet_list += item

    term = nodes.term(text="Parameters")
    item = nodes.definition_list_item("", term, bullet_list)

    definition_list = nodes.definition_list()
    definition_list += item

    return definition_list


def create_example(doc: dict) -> nodes.literal_block | None:
    example = doc.get("example", None)
    if example is None:
        return None

    text = example.strip()
    block = nodes.literal_block(text=text, language="yaml")
    return block


def setup(app):
    app.add_directive("cauliflow-filters", CauliFlowFilterDirective)
    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



from cauliflow.context import init_flowdata
from cauliflow.node import ProcessNode, node


@node.register("root")
class RootNode(ProcessNode):
    async def process(self) -> None:
        init_flowdata()



from pathlib import Path

import aiofiles

from cauliflow.node import ArgSpec, ProcessNode, node


@node.register("out_file")
class OutFileNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Output the data to the file.
      description:
        - Output the data to the file.
      parameters:
        path:
          description:
            - The path to the file to output.
        src:
          description:
            - Data to output.
    EXAMPLE: |-
      # Output hello to test.txt
      - out_file:
          name: "out"
          path: "./test.txt"
          src: "hello"

      # Output csv data in blackboard to test.txt
      - out_file:
          name: "out"
          path: "./test.txt"
          src: "{{ bb.csv }}"
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        return {
            "path": ArgSpec(type="path", required=True),
            "src": ArgSpec(type="str", required=True),
        }

    async def process(self) -> None:
        path = Path(self.params["path"])
        src = self.params["src"]
        async with aiofiles.open(path, mode="a") as f:
            await f.write(src + "\n")



import pytest
from cauliflow.flowdata import FlowData


def test_flowdata():
    d = FlowData()

    d["a"] = 1
    assert d["a"] == 1

    with pytest.raises(KeyError):
        d["a"] = 1

    d["b"] = 2

    assert d["a"] == 1
    assert d["b"] == 2



from pathlib import Path

import pytest

from cauliflow.context import ctx_blackboard, ctx_flowdata, ctx_macros
from cauliflow.loader import flow_from_yaml


@pytest.mark.asyncio
async def test_sequential(request, init_plugins):
    current_dir = Path(request.fspath).parent
    path = current_dir / "flows/sequential.yml"
    flows = flow_from_yaml(path)
    await flows.run()

    flowdata = ctx_flowdata.get()
    blackboard = ctx_blackboard.get()
    assert blackboard["add1"] == 11
    assert flowdata["add2"] == 12


@pytest.mark.asyncio
async def test_macro(request, init_plugins):
    current_dir = Path(request.fspath).parent
    path = current_dir / "flows/sequential.yml"
    flows = flow_from_yaml(path)

    mcr = ctx_macros.get()
    mcr["init_val"] = 100
    ctx_macros.set(mcr)

    await flows.run()

    flowdata = ctx_flowdata.get()
    blackboard = ctx_blackboard.get()
    assert blackboard["add1"] == 101
    assert flowdata["add2"] == 102


@pytest.mark.asyncio
async def test_concurrent(request, init_plugins):
    current_dir = Path(request.fspath).parent
    path = current_dir / "flows/concurrent.yml"
    flows = flow_from_yaml(path)
    await flows.run()

    blackboard = ctx_blackboard.get()
    assert blackboard["add1"] == 11
    assert blackboard["add2"] == 21


@pytest.mark.asyncio
async def test_composite(request, init_plugins):
    current_dir = Path(request.fspath).parent
    path = current_dir / "flows/composite.yml"
    flows = flow_from_yaml(path)
    await flows.run()

    blackboard = ctx_blackboard.get()
    assert blackboard["add2"] == 12
    assert blackboard["add4"] == 22


@pytest.mark.asyncio
async def test_flowonly(request, init_plugins):
    current_dir = Path(request.fspath).parent
    path = current_dir / "flows/flowonly.yml"
    flows = flow_from_yaml(path)
    await flows.run()

    blackboard = ctx_blackboard.get()
    assert blackboard["add2"] == 12



import time
from typing import Any


def _str(target: Any) -> str:
    """
    description:
      - Convert the value into a string.
    """
    return str(target)


def _int(target: Any) -> int:
    """
    description:
      - Convert the value into an integer.
    """
    return int(target)


def _float(target: Any) -> float:
    """
    description:
      - Convert the value into a float.
    """
    return float(target)


def _bool(target: Any) -> bool:
    """
    description:
      - Convert the value into a boolean.
    """
    return bool(target)


def dict_keys(target: dict) -> list:
    """
    description:
      - Return a list of all the keys from the dictionary.
    example: |-
      # item: ["key1", "key2"]
      item: "{{ {'key1': 1, 'key2': 2} | dict_keys }}"
    """
    return list(target.keys())


def dict_values(target: dict) -> list:
    """
    description:
      - Return a list of all the values from the dictionary.
    example: |-
      # item: [1, 2]
      item: "{{ {'key1': 1, 'key2': 2} | dict_values }}"
    """
    return list(target.values())


def dict2item(key_name: str, val_name: str, target: dict | None = None) -> list[dict]:
    """
    description:
      - Return a list of dictionaries from the dictionary.
    parameters:
      key_name:
        description: The name of the dictionary’s keys.
      val_name:
        description: The name of the dictionary’s values.
    example: |-
      # item: [{"k": "key1", "v": 1}, {"k": "key2", "v": 2}]
      item: "{{ {'key1': 1, 'key2': 2} | dict2item('k','v') }}"
    """
    if not isinstance(target, dict):
        raise ValueError
    dictlist = []
    for k, v in target.items():
        dictlist.append({key_name: k, val_name: v})
    return dictlist


def str_pvts(tstamp: float) -> str:
    """
    description:
      - Return a string of formated timestamp from the float value of pyepics timestamp.
    example: |-
      # item: "2025-05-26 14:48:29.99555"
      item: "{{ 1748238509.995554 | str_pvts }}"
    """
    tstamp, frac = divmod(tstamp, 1)
    return "%s.%5.5i" % (
        time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(tstamp)),
        round(1.0e5 * frac),
    )


def join(separator: str, target: list[str] | None = None) -> str:
    """
    description:
      - Join a list of string with separator.
    parameters:
      separator:
        description: Separator for the concatenation of the strings.
    example: |-
      # item: "Hello, world"
      item: "{{ ["Hello", "world"] | join(', ') }}"
    """
    if not isinstance(target, list):
        raise ValueError
    return separator.join(target)


FILTERS = {
    "str": _str,
    "int": _int,
    "float": _float,
    "bool": _bool,
    "str_pvts": str_pvts,
    "dict_keys": dict_keys,
    "dict_values": dict_values,
    "dict2item": dict2item,
    "join": join,
}



from cauliflow.node import ArgSpec, ProcessNode, node


@node.register("message")
class MessageNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Output data to flowdata or blackboard.
      description:
          - This node outputs data to flowdata or blackboard
      parameters:
        msg:
          description:
          - Output data.
    EXAMPLE: |-
      # Output "hello" to the flowdata["msg"]
      - message:
          name: "msg"
          msg: "hello"

      # Output the value of 'foo' field in the blackboard to the flowdata["msg"]
      - message:
          name: "msg"
          msg: "{{ bb.foo }}"

      # Output "hello" to the flowdata["field"]
      - message:
          name: "msg"
          msg: "hello"
          out_field: "field"

      # Output "hello" to the blackboard["msg"]
      - message:
          name: "msg"
          msg: "hello"
          out_bb: yes
    """

    async def process(self) -> None:
        self.output(self.params["msg"])

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {"msg": ArgSpec(type="any", required=True)}



from functools import singledispatchmethod

from cauliflow.logging import get_logger
from cauliflow.node import ArgSpec, ProcessNode, node
from cauliflow.variable import Variable

_logger = get_logger(__name__)


@node.register("for_list")
class ForList(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Loop the list of list or dict and create a new list to output.
      description:
        - Loop the list of list or dict and create a new list to output.
      parameters:
        lists:
          description:
            - List of list or dict to make a nested loop.
            - Single list or dict can be passed.
        expression:
          description:
            - Expression of a item for a created list.
            - Item in the loop of list can be referred as item*.
            - Item in the loop of dict can be referred as item*_val and item*_key.
            - For the first loop item can be referred as item0.
            - For the second loop item can be referred as item1.
        filter:
          description:
            - Condition not to add item for a crated list.
            - Loop item can be refereed as same as in expression.
    EXAMPLE: |-
      # Create a list from list of list with filter parameter.
      # Output: {'for_list': [3, 4, 6]}
      - for_list:
          name: "for_list"
          lists: [[1, 2], [3, 4]]
          expression: "item0*item1"
          filter: "item0*item1>7"

      # Create a list from list of list and dict.
      # Output: {'for_list': ['head1:val1', 'head1:val2', 'head2:val1', 'head2:val2']}
      - for_list:
          name: "for_list"
          lists: [["head1", "head2"], {"name1": "val1", "name2": "val2"}]
          expression: "item0 + ':' + item1_val"

      # Create a list from a single list.
      # Output: {'for_list': [2, 4, 6]}
      - for_list:
          name: "for_list"
          lists: [1, 2, 3, 4]
          expression: "item0*2"
          filter: "item0*2>7"

      # Create a list from a dict.
      # Output: {'for_list': ["key1-1", "key2-2"]}
      - for_list:
          name: "for_list"
          lists: {"key1": 1, "key2": 2}
          expression: "item0_key + '-' + item0_val | str"

      # Create a list from nested lists of a dict.
      # Output: {'for_list': ["hello, world", "foobar"]}
      - for_list:
          name: "for_list"
          lists: [[{"first": "hello, ", "second": "world"}, {"first": "foo", "second": "bar"}]]
          expression:  "item0.first + item0.second"
    """

    def __init__(self, name: str, param_dict: dict | None = None):
        super().__init__(name, param_dict)
        self.variable = None
        self.filter = None

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {
            "lists": ArgSpec(type="list[list|dict]", required=True),
            "expression": ArgSpec(type="expression", required=True),
            "filter": ArgSpec(type="expression", required=False, default=None),
        }

    async def process(self) -> None:
        if self.variable is None:
            self.variable = Variable("{{" + self.params["expression"] + "}}")
        if self.filter is None and self.params["filter"] is not None:
            self.filter = Variable("{{" + self.params["filter"] + "}}")

        lists = self.params["lists"]
        is_dict = isinstance(lists, dict)

        if is_dict:
            items = self._for_loop(lists, 0, [], {})
            self.output(items)
            return

        if not isinstance(lists, list):
            raise ValueError

        if len(lists) < 1:
            return

        has_list_or_dict = isinstance(lists[0], (list, dict))
        loop_lists = lists if has_list_or_dict else [lists]

        items = self._for_loop(loop_lists[0], 0, loop_lists[1:], {})
        self.output(items)

    @singledispatchmethod
    def _for_loop(
        self, iterable, depth: int, remaining_lists: list, var_dict: dict
    ) -> list:
        _logger.critical("Input must be a list or dict")
        return []

    @_for_loop.register
    def _(
        self, for_list: list, depth: int, remaining_lists: list, var_dict: dict
    ) -> list:
        self._validate_variable()

        results = []
        for item in for_list:
            var_dict[f"item{depth}"] = item
            results.extend(self._process_next_level(depth, remaining_lists, var_dict))
        return results

    @_for_loop.register
    def _(
        self, for_dict: dict, depth: int, remaining_lists: list, var_dict: dict
    ) -> list:
        self._validate_variable()

        results = []
        for key, value in for_dict.items():
            var_dict[f"item{depth}_key"] = key
            var_dict[f"item{depth}_val"] = value
            results.extend(self._process_next_level(depth, remaining_lists, var_dict))
        return results

    def _process_next_level(
        self, depth: int, remaining_lists: list, var_dict: dict
    ) -> list:
        if remaining_lists:
            return self._for_loop(
                remaining_lists[0], depth + 1, remaining_lists[1:], var_dict
            )
        if self.filter and self.filter.fetch(extend=var_dict):
            return []
        return [self.variable.fetch(extend=var_dict)]  # type: ignore

    def _validate_variable(self):
        if self.variable is None:
            raise ValueError("Variable is not set.")


@node.register("for_dict")
class ForDict(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Loop the list of list or dict and create a new dict to output.
      description:
        - Loop the list of list or dict and create a new dict to output.
      parameters:
        lists:
          description:
            - List of list or dict to make a nested loop.
            - Single list or dict can be passed.
        key:
          description:
            - Expression of a item of key for a created dict.
            - Item in the loop of list can be referred as item*.
            - Item in the loop of dict can be referred as item*_val and item*_key.
            - For the first loop item can be referred as item0.
            - For the second loop item can be referred as item1.
        val:
          description:
            - Expression of of a item of value for a created dict.
            - Loop item can be refereed as same as in key.
        filter:
          description:
            - Condition not to add item for a crated dict.
            - Loop item can be refereed as same as in expression.
    EXAMPLE: |-
      # Create a dict from list of list and dict.
      # Output: {'for_dict': {'head1:name1': 'val1', 'head1:name2': 'val2', 'head2:name1': 'val1', 'head2:name2': 'val2'}}
      - for_dict:
          name: "for_dict"
          lists: [["head1", "head2"], {"name1": "val1", "name2": "val2"}]
          key: "item0 + ':' + item1_key"
          val: "item1_val"

      # Create a dict from a single list.
      # Output: {'for_dict': {"key1": 2, "key2": 4, "key3": 6}}
      - for_dict:
          name: "for_dict"
          lists: [1, 2, 3, 4]
          key: "'key' + item0 |str"
          val: "item0*2"
          filter: "item0*2>7"

      # Create a dict from a dict.
      # Output: {'for_dict': {"key1-1": 1, "key2-2": 2}}
      - for_dict:
          name: "for_dict"
          lists: {"key1": 1, "key2": 2}
          key: "item0_key + '-' + item0_val | str"
          val: "item0_val"
    """

    def __init__(self, name: str, param_dict: dict | None = None):
        super().__init__(name, param_dict)
        self.key = None
        self.val = None
        self.filter = None

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {
            "lists": ArgSpec(type="list", required=True),
            "key": ArgSpec(type="expression", required=True),
            "val": ArgSpec(type="expression", required=True),
            "filter": ArgSpec(type="expression", required=False, default=None),
        }

    async def process(self) -> None:
        if self.key is None:
            self.key = Variable("{{" + self.params["key"] + "}}")
        if self.val is None:
            self.val = Variable("{{" + self.params["val"] + "}}")
        if self.filter is None and self.params["filter"] is not None:
            self.filter = Variable("{{" + self.params["filter"] + "}}")

        lists = self.params["lists"]
        is_dict = isinstance(lists, dict)

        if is_dict:
            items = self._for_loop(lists, 0, [], {})
            self.output(items)
            return

        if not isinstance(lists, list):
            raise ValueError

        if len(lists) < 1:
            return

        has_list_or_dict = isinstance(lists[0], (list, dict))
        loop_lists = lists if has_list_or_dict else [lists]

        items = self._for_loop(loop_lists[0], 0, loop_lists[1:], {})
        self.output(items)

    @singledispatchmethod
    def _for_loop(
        self, iterable, depth: int, remaining_lists: list, var_dict: dict
    ) -> dict:
        _logger.critical("Input must be a list or dict")
        return {}

    @_for_loop.register
    def _(
        self, for_list: list, depth: int, remaining_lists: list, var_dict: dict
    ) -> dict:
        self._validate_variable()

        results = {}
        for item in for_list:
            var_dict[f"item{depth}"] = item
            results.update(self._process_next_level(depth, remaining_lists, var_dict))
        return results

    @_for_loop.register
    def _(
        self, for_dict: dict, depth: int, remaining_lists: list, var_dict: dict
    ) -> dict:
        self._validate_variable()

        results = {}
        for key, value in for_dict.items():
            var_dict[f"item{depth}_key"] = key
            var_dict[f"item{depth}_val"] = value
            results.update(self._process_next_level(depth, remaining_lists, var_dict))
        return results

    def _process_next_level(
        self, depth: int, remaining_lists: list, var_dict: dict
    ) -> dict:
        if remaining_lists:
            return self._for_loop(
                remaining_lists[0], depth + 1, remaining_lists[1:], var_dict
            )
        if self.filter and self.filter.fetch(extend=var_dict):
            return {}
        _key = self.key.fetch(extend=var_dict)  # type: ignore
        _val = self.val.fetch(extend=var_dict)  # type: ignore
        return {_key: _val}

    def _validate_variable(self):
        if self.key is None or self.val is None:
            raise ValueError("Variable is not set.")



import pytest
from cauliflow.context import ctx_flowdata
from cauliflow.plugins.message import MessageNode


@pytest.mark.asyncio
async def test_message_node(init_context_vars):
    node = MessageNode(name="node", param_dict={"msg": "test"})
    await node.run()
    flowdata = ctx_flowdata.get()
    message = flowdata["node"]
    assert message == "test"



import csv
from enum import StrEnum
from pathlib import Path

from cauliflow.logging import get_logger
from cauliflow.node import ArgSpec, ProcessNode, node

_logger = get_logger(__name__)


class DataFormat(StrEnum):
    ARRAY = "array"
    DICT = "dict"
    KEYVALUE = "key_value"


@node.register("in_csv")
class InputCSVNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Read csv file and output the data to flowdata or blackboard.
      description:
        - Read csv file and output the data to flowdata or blackboard.
        - Output format can be chosen array, dict or key_value.
      parameters:
        path:
          description:
            - The path to the csv file to read.
        format:
          description:
            - Output format.
            - Choose array, dict or key_value.
    EXAMPLE: |-
      # Assume following csv file is located at "./file.csv"
      # id, name
      # foo, John
      # bar, Tom

      # Read the csv file as key_value and output it to blackboard
      # Output: {"csv": {'foo': 'John', 'bar': 'Tom'}}
      - in_csv:
          name: "csv"
          path: "./file.csv"
          format: "key_value"
          out_bb: yes

      # Read the csv file as dict
      # Output: {"csv": [{'id': 'foo', 'name': 'John'}, {'id': 'bar', 'name': 'Tom'}]}
      - in_csv:
          name: "csv"
          path: "./file.csv"
          format: "dict"

      # Read the csv file as dict
      # Output: {"csv": [["id", "name"], ["foo", "John"], ["bar", "Tom"]]}
      - in_csv:
          name: "csv"
          path: "./file.csv"
          format: "array"
    """

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {
            "path": ArgSpec(type="path", required=True),
            "format": ArgSpec(type="str", required=False, default="key_value"),
        }

    async def process(self) -> None:
        path = self.params["path"]
        format = self.params["format"]

        csvdata = self.get_csvdata(Path(path), format)
        self.output(csvdata)

    def get_csvdata(self, path: Path, format: str | DataFormat) -> dict | list | None:
        with path.open(newline="") as csvfile:
            if format == DataFormat.DICT:
                reader = csv.DictReader(csvfile, skipinitialspace=True)
                dictlist = [row for row in reader]
                return dictlist

            reader = csv.reader(csvfile, skipinitialspace=True)
            if format == DataFormat.ARRAY:
                csvdata = [row for row in reader]
                return csvdata

            if format == DataFormat.KEYVALUE:
                csvdata = {}
                for row in reader:
                    csvdata[row[0]] = row[1]
                return csvdata

            _logger.warning(f"format:{format} is not matched")



from contextvars import ContextVar
from dataclasses import dataclass

from cauliflow.blackboard import BlackBoard
from cauliflow.flowdata import FlowData
from cauliflow.macros import Macros


@dataclass
class ContextFlows:
    debug: bool = False


@dataclass
class ContextFlow:
    name: str | None = None


@dataclass
class ContextNode:
    name: str | None = None


ctx_macros = ContextVar("cauliflow.ctx_macros", default=Macros())
ctx_flowdata = ContextVar("cauliflow.ctx_flowdata", default=FlowData())
ctx_blackboard = ContextVar("cauliflow.ctx_blackboard", default=BlackBoard())
ctx_flows = ContextVar("pypelind.ctx_flows", default=ContextFlows())
ctx_flow = ContextVar("pypelind.ctx_flow", default=ContextFlow())
ctx_node = ContextVar("pypelind.ctx_node", default=ContextNode())


def init_flowdata() -> None:
    ctx_flowdata.set(FlowData())



import asyncio
from abc import ABC, abstractmethod

from cauliflow.context import (
    ContextFlow,
    ContextNode,
    ctx_blackboard,
    ctx_flow,
    ctx_node,
)
from cauliflow.logging import get_logger
from cauliflow.node import Node, node

_logger = get_logger(__name__)


class Flow:
    def __init__(self, name: str | None = None):
        self.blackboard = ctx_blackboard.get()
        self.name = name
        self.root: Node = node.create("root", name="root", param_dict={})
        self.nodes: dict[str, Node] = {}
        self.nodes["root"] = self.root

    async def run(self) -> None:
        ctx_flow.set(ContextFlow(name=self.name))
        _logger.debug("run starts")
        await self.root.run()
        ctx_node.set(ContextNode(name=None))
        _logger.debug("run end")

    def create_node(
        self, _plugin_name: str, _parent: str, name: str, param_dict: dict
    ) -> None:
        if name in self.nodes:
            raise KeyError(
                f"The node name is in conflict: {name}. Please change the node name."
            )

        self.nodes[name] = node.create(_plugin_name, name=name, param_dict=param_dict)

        if not _parent:
            return

        if "." not in _parent:
            self.nodes[_parent].add_child(self.nodes[name])
            return

        parent, field = _parent.split(".")
        self.nodes[parent].add_child(self.nodes[name], field)


class Flows(ABC):
    @abstractmethod
    async def run(self): ...

    def __init__(self):
        self.flows: list[Flow | Flows] = []

    def append(self, flow: "Flow | Flows"):
        self.flows.append(flow)

    def extend(self, flow: list["Flow | Flows"]):
        self.flows.extend(flow)


class ConcurrentFlows(Flows):
    async def run(self) -> None:
        async with asyncio.TaskGroup() as tg:
            for task in self.flows:
                tg.create_task(task.run())


class SequentialFlows(Flows):
    async def run(self) -> None:
        for task in self.flows:
            await task.run()



import asyncio
from enum import StrEnum
from typing import Any, TypedDict

from aiohttp import ClientResponse, ClientSession, ClientTimeout

from cauliflow.node import ArgSpec, ProcessNode, node


class ResDataFormat(StrEnum):
    JSON = "json"
    TEXT = "text"


class MethodType(StrEnum):
    GET = "get"
    PUT = "put"
    POST = "post"
    PATCH = "patch"
    DELETE = "delete"


class SuccessOut(TypedDict):
    status: int
    data: Any


@node.register("http")
class HTTPNode(ProcessNode):
    """
    DOCUMENTATION:
      short_description: Send a request to a HTTP endpoint.
      description:
          - Send a request to a HTTP endpoint.
      parameters:
        url:
          description:
            - URL.
        method:
          description:
            - "HTTP method. Following methods are available: get, put, post, patch, and delete."
        timeout:
          description:
            - Timeout in second.
        body:
          description:
            - Request body. This parameter is used for PUT, POST, and PATCH method.
        format:
          description:
            - Data format to decode a response body. It can be either text or json.
    EXAMPLE: |-
      # Get with json format
      # Output: {"http": {"data": {"foo": "bar"}, "status": 200}}
      - http:
          name: "http"
          url: "http://example.com/api/data"
          format: "json"

      # Get with text format
      # Output: {"http": {"data": '{"foo": "bar"}', "status": 200}}
      - http:
          name: "http"
          url: "http://example.com/api/data"
          format: "json"

      # Timeout Error
      # Output: {"http": '{"error": "TimeoutError"}'}
      - http:
          name: "http"
          url: "http://example.com/api/data"
          format: "json"
          timeout: "1"
    """

    async def process(self) -> None:
        timeout = ClientTimeout(total=self.params["timeout"])
        method = self.params["method"]

        try:
            async with ClientSession(timeout=timeout) as session:
                match method:
                    case MethodType.GET:
                        out = await self._get(session)
                    case MethodType.PUT:
                        out = await self._put(session)
                    case MethodType.POST:
                        out = await self._post(session)
                    case MethodType.PATCH:
                        out = await self._patch(session)
                    case MethodType.DELETE:
                        out = await self._delete(session)
                    case _:
                        out = {"error": "Illegal HTTP method"}
                self.output(out)
        except asyncio.TimeoutError:
            out = {"error": "TimeoutError"}
            self.output(out)

    def set_argument_spec(self) -> dict[str, ArgSpec]:
        self.set_common_output_args()
        return {
            "url": ArgSpec(type="str", required=True),
            "format": ArgSpec(type="str", required=False, default="text"),
            "timeout": ArgSpec(type="float", required=False, default=10),
            "method": ArgSpec(type="str", required=False, default="get"),
            "body": ArgSpec(type="str", required=False, default=""),
        }

    async def _get(self, session: ClientSession) -> SuccessOut:
        async with session.get(self.params["url"]) as resp:
            out = await self._get_output(resp)

        return out

    async def _put(self, session: ClientSession) -> SuccessOut:
        async with session.put(self.params["url"], data=self.params["body"]) as resp:
            out = await self._get_output(resp)

        return out

    async def _post(self, session: ClientSession) -> SuccessOut:
        async with session.post(self.params["url"], data=self.params["body"]) as resp:
            out = await self._get_output(resp)

        return out

    async def _patch(self, session: ClientSession) -> SuccessOut:
        async with session.patch(self.params["url"], data=self.params["body"]) as resp:
            out = await self._get_output(resp)

        return out

    async def _delete(self, session: ClientSession) -> SuccessOut:
        async with session.delete(self.params["url"]) as resp:
            out = await self._get_output(resp)

        return out

    async def _get_output(self, resp: ClientResponse) -> SuccessOut:
        if self.params["format"] == ResDataFormat.JSON:
            data = await resp.json()
        else:
            data = await resp.text()
        out: SuccessOut = {"status": resp.status, "data": data}
        return out



import logging

from cauliflow.context import ctx_flow, ctx_node


class InjectingFilter(logging.Filter):
    def filter(self, record):
        flow = ctx_flow.get()
        node = ctx_node.get()
        record.flow_name = flow.name
        record.node_name = node.name
        return True


default_handler = logging.StreamHandler()
default_handler.setFormatter(
    logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(flow_name)s - %(node_name)s - %(message)s"
    )
)

default_handler.addFilter(InjectingFilter())


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)

    if not logger.hasHandlers():
        logger.addHandler(default_handler)

    return logger



from functools import partial
from operator import contains
from typing import Any

from lark import Lark, ParseTree, Transformer, Tree, v_args

from cauliflow.context import ctx_blackboard, ctx_flowdata, ctx_macros
from cauliflow.filters import FILTERS

_grammar = r"""
    start: (text | expression_wrapper)*
    text: /[^{}]+/
    expression_wrapper: "{{" expression "}}"
    ?expression: disjunction

    ?disjunction: conjunction
        | conjunction "or" conjunction -> or_

    ?conjunction: inversion
        | inversion "and" inversion -> and_

    ?inversion: comparison
        | "not" inversion -> not_

    ?comparison: sum
        | comparison "<" sum        -> lt
        | comparison ">" sum        -> gt
        | comparison "==" sum       -> eq
        | comparison "!=" sum       -> ne
        | comparison ">=" sum       -> ge
        | comparison "<=" sum       -> le
        | comparison "in" sum       -> contains_
        | comparison "not in" sum   -> not_contains
        | comparison "is" sum       -> is_
        | comparison "is not" sum   -> is_not

    ?sum: product
        | sum "+" product   -> add
        | sum "-" product   -> sub

    ?product: atom_expr
        | product "*" atom_expr  -> mul
        | product "/" atom_expr  -> div
        | product "//" atom_expr -> floor
        | product "%" atom_expr  -> mod

    ?atom_expr: atom
        | atom_expr "[" index "]" -> getitem
        | atom_expr "." NAME      -> getitem
        | atom_expr "|" filter    -> filter

    filter: NAME [arguments_wrapper] -> filter_func

    arguments_wrapper: "(" [arguments] ")" -> arguments_wrapper
    arguments: [arg ("," arg)*] -> arguments
    arg: expression -> arg

    ?index: expression
        | slice

    slice: expression ":" expression

    ?atom: number
        | "-" atom         -> neg
        | NAME             -> var
        | "True"           -> true
        | "False"          -> false
        | "None"           -> none
        | string
        | list
        | dict
        | "(" disjunction ")"

    ?number: INT    -> integer
        | FLOAT     -> float

    string: STRING

    list: "[" [disjunction ("," disjunction)*] "]"
    dict: "{" [pair ("," pair)*] "}"
    pair: string ":" disjunction

    STRING: /([ubf]?r?|r[ubf])("(?!"")[\s\S]*?(?<!\\)(\\\\)*?"|'(?!'')[\s\S]*?(?<!\\)(\\\\)*?')/i

    %import common.CNAME -> NAME
    %import common.INT
    %import common.FLOAT
    %import common.NUMBER
    %import common.WS_INLINE
    %ignore WS_INLINE
"""


@v_args(inline=True)  # Affects the signatures of the methods
class OperatorTree(Transformer):
    from operator import (
        add,
        and_,
        eq,
        ge,
        getitem,
        gt,
        is_,
        is_not,
        le,
        lt,
        mod,
        mul,
        ne,
        neg,
        not_,
        or_,
        sub,
    )
    from operator import floordiv as floor
    from operator import truediv as div

    float = float
    integer = int
    slice = slice

    def __init__(self, vars: dict = {}):
        self.vars = vars

    def start(self, *args) -> str:
        if len(args) > 1:
            return "".join(args)
        if len(args) < 1:
            return ""
        return args[0]

    def string(self, s: str) -> str:
        return s[1:-1]

    def none(self) -> None:
        return None

    def true(self) -> bool:
        return True

    def false(self) -> bool:
        return False

    def list(self, *args) -> list:
        return list(args)

    def pair(self, key, value) -> tuple:
        return (key, value)

    def dict(self, *args) -> dict:
        return dict(args)

    def text(self, string) -> str:
        return str(string)

    def expression_wrapper(self, expression) -> Any:
        return expression

    def contains_(self, a: Any, b: Any) -> bool:
        return contains(b, a)

    def not_contains(self, a: Any, b: Any) -> bool:
        return not contains(b, a)

    def filter(self, a: Any, f: Any) -> Any:
        return f(a)

    def filter_func(self, name: Any, arguments: Any) -> Any:
        filters = FILTERS
        if name not in filters:
            raise KeyError(f"{name} is not a valid filter")
        f = filters[name]
        if arguments is not None:
            f = partial(f, *arguments)
        return f

    def arguments_wrapper(self, args) -> Any:
        return args

    def arguments(self, *args) -> Any:
        return list(args)

    def arg(self, a) -> Any:
        return a

    def var(self, name: str) -> Any:
        try:
            return self.vars[name]
        except KeyError:
            raise Exception("Variable not found: %s" % name)


_parser = Lark(_grammar, start="start", parser="lalr")


class Variable:
    def __init__(self, expression: Any):
        self.expression = expression
        self.parse_tree = None
        self.val = None
        self.has_var = None

        if isinstance(expression, str):
            self.parse_tree = self._compile(expression)
            self.has_var = self._find_var(self.parse_tree)

        if not self.has_var:
            transformer = OperatorTree({})
            self.val = transformer.transform(self.parse_tree)

    def _compile(self, expression: str) -> ParseTree:
        return _parser.parse(expression)

    def fetch(self, extend: dict = {}) -> Any:
        if self.parse_tree is None:
            return self.expression

        if not self.has_var:
            return self.val

        bb = ctx_blackboard.get()
        fd = ctx_flowdata.get()
        mcr = ctx_macros.get()
        vars = {"bb": bb, "fd": fd, "macro": mcr}
        vars.update(extend)
        transformer = OperatorTree(vars)
        result = transformer.transform(self.parse_tree)
        return result

    def _find_var(self, tree: Tree):
        for subtree in tree.iter_subtrees():
            if subtree.data == "var":
                return True
        return False



import pytest

from cauliflow.blackboard import BlackBoard
from cauliflow.context import ctx_blackboard, ctx_flowdata, ctx_macros
from cauliflow.flowdata import FlowData
from cauliflow.macros import Macros
from cauliflow.variable import Variable


@pytest.fixture
def context_vars():
    bb = BlackBoard({"bb1": "foo", "dict": {"foo": "bar"}})
    fd = FlowData({"fd1": "bar"})
    macros = Macros({"mc1": "foobar"})

    ctx_blackboard.set(bb)
    ctx_flowdata.set(fd)
    ctx_macros.set(macros)


@pytest.mark.parametrize(
    "input,expected",
    [
        # Basic expression
        ("", ""),
        ("''", "''"),
        ("{{ -1 }}", -1),
        ("This is a test.", "This is a test."),
        ("{{ [1, 2, 3] }}", [1, 2, 3]),
        ("{{ {'key': 'val'} }}", {"key": "val"}),
        ("{{ 'foo' + 'bar' }}", "foobar"),
        ("{{ 'foo\t' + 'bar' }}", "foo\tbar"),
        ("{{ 'foo\n' + 'bar' }}", "foo\nbar"),
        ("{{ True or False }}", True),
        ("{{ True and False }}", False),
        ("{{ not True }}", False),
        ("{{ 1 > 0 }}", True),
        ("{{ 1 < 0 }}", False),
        ("{{ 1 == 1 }}", True),
        ("{{ 1 != 1 }}", False),
        ("{{ 1 >= 1 }}", True),
        ("{{ 1 <= 1 }}", True),
        ("{{ None is None }}", True),
        ("{{ None is not None }}", False),
        ("{{ (1+1)*(2-1)/2 }}", 1),
        ("{{ 0.1 + 0.1 }}", 0.2),
        ("{{ 5//3 }}", 1),
        ("{{ 5%3 }}", 2),
        # Filters
        ("{{ 'test' + 1 | str }}", "test1"),
        ("{{ '1' | float }}", 1),
        ("{{ '1' | float | int | str }}", "1"),
        ("{{ {'key1': 1, 'key2': 2} | dict_keys() }}", ["key1", "key2"]),
        ("{{ {'key1': 1, 'key2': 2} | dict_values }}", [1, 2]),
        (
            "{{ {'key1': 1, 'key2': 2} | dict2item('k','v') }}",
            [{"k": "key1", "v": 1}, {"k": "key2", "v": 2}],
        ),
        # Variable
        ("{{ 'bb1' in bb }}", True),
        ("{{ 'bb1' not in bb }}", False),
        ("{{ bb['bb1'] + fd['fd1'] + '-' + macro['mc1'] }}", "foobar-foobar"),
        ("{{ bb.bb1 + fd.fd1 + '-' + macro.mc1 }}", "foobar-foobar"),
        ("{{ bb['bb1'] }}bar-{{ macro['mc1'] }}", "foobar-foobar"),
        ("{{ bb['bb1'] }} bar-  {{ macro['mc1'] }}", "foobar-  foobar"),
        ("{{ bb['bb1'] + ' ' }} bar-  {{ macro['mc1'] }}", "foo bar-  foobar"),
        ("{{ bb['dict'] | dict_keys }}", ["foo"]),
        ("{{ bb.dict.foo }}", "bar"),
        ("{{ bb.dict['foo'] }}", "bar"),
    ],
)
def test_variable(context_vars, input, expected):
    var = Variable(input)
    data = var.fetch()
    assert data == expected
